# -*- coding: utf-8 -*-
from PyAstronomy.pyaC import ImportCheck
from PyAstronomy.pyaC import pyaErrors as PE

ic = ImportCheck(["numpy", "scipy"], required=["numpy"]) 

from .periodBase import *
from .lombScargle import *
from .gls import *

if ic.check["scipy"]:
  from .fourier import *
else:
  PE.warn(PE.PyARequiredImport("No Fourier transform will be available, because scipy cannot be imported.", \
                               where="Import of PyPeriod", solution="Install scipy."))



import unittest
class TestSanityOfNotebook(unittest.TestCase):
    def setUp(self):
        pass
    def tearDown(self):
        pass
        
    def testsanity_example(self):
        #!/usr/bin/env python
        # coding: utf-8
        
        # Custom line with additional jitter term
        # ========================================
        
        # In[1]:
        
        
        import numpy as np
        import matplotlib.pylab as plt
        from PyAstronomy import funcFit2 as fuf2
        import scipy.optimize as sco
        import matplotlib
        matplotlib.use('Agg')
        
        np.random.seed(1234)
        
        class LinMod(fuf2.MBO):
            """ Linear model with additional jitter """
        
            def __init__(self):
                fuf2.MBO.__init__(self, pars=["const", "slope", "jitter"], rootName="LinMod")
        
            def evaluate(self, x):
                """ Evaluate model """
                return self["const"] + x * self["slope"]
        
            def logL(self, x, y, yerr, **kwargs):
                """ ln(Likelihood) including jitter as additional term """
                yr = np.sqrt(yerr**2 + self["jitter"]**2)
                m = self.evaluate(x)
                lnl = -len(x)/2.0*np.log(2.*np.pi) - np.sum(np.log(yr)) - 0.5 * np.sum((m-y)**2/(yr**2))
                return lnl
        
        # Instantiate model
        lm = LinMod()
        # Starting values
        lm["slope"] = 1.1
        lm["const"] = -0.5
        lm["jitter"] = 1
        
        # Use -log(L) as SciPy-like objective function and call it logl
        lm.addSPLikeObjf("-logl", "logl")
        
        # Get some 'data' and add Gaussian noise with STD 1+3**2=10
        x = np.arange(150.)
        y = lm.evaluate(x) + np.random.normal(0,np.sqrt(1+3**2),len(x))
        # Nominal error has STD 1
        yerr = np.ones_like(x)
        
        lm.thaw(["slope", "const", "jitter"])
        
        fr = sco.fmin_powell(lm.logl, x0=lm.freeParamVals(), args=(x,y,yerr))
        lm.setFreeParamVals(fr)
        
        lm.parameterSummary()
        
        plt.errorbar(x, y, yerr=yerr, fmt='b+')
        plt.plot(x, lm.evaluate(x), 'r--')
        plt.show()
        



from __future__ import print_function, division
import unittest

class SanityOfBaraffe98Tracks(unittest.TestCase):
  
  def setUp(self):
    pass
  
  def tearDown(self):
    pass
  
  def sanity_example(self):
    """
      Checking sanity of Baraffe tracks (1998) example
    """
    from PyAstronomy.pyasl import resBased as rb
    import matplotlib.pylab as plt
    import numpy as np
    
    bt = rb.Baraffe98Tracks()
    
    print("Unique metallicity values: ", bt.getUniqueValues("Met"))
    print("Unique Y values: ", bt.getUniqueValues("Y"))
    print("Unique Lmix values: ", bt.getUniqueValues("Lmix"))
    print("Unique mass values: ", bt.getUniqueValues("Mass"))
    
    # Get model data and plot log10(age) versus effective temperature
    m = bt.getModelData((0.0, 0.275, 1.0, 0.3))
    plt.plot(np.log10(m.Age*1e9), m.Teff, 'b.-')
    
    # Find all models with metallicity 0.0, Y 0.275, Lmix 0.0,
    # and any mass.
    models = bt.findModels(Met=0.0, Mass=None, Y=0.275, Lmix=1.0)
    # Print out the model parameters.
    print()
    print("Number of models found: ", len(models))
    for i, model in enumerate(models):
      print("Model no. %3d : Met = %3.1f, Y = %5.3f, Lmix = %3.1f, Mass = %4.2f" \
            % ((i+1,) + model))
    
    # Finally, show the plot
#     plt.show()


class SanityOfNasaExoplanetArchive(unittest.TestCase):
  
  def setUp(self):
    pass
  
  def tearDown(self):
    pass
  
  def sanity_example(self):
    """
      Checking sanity of exoplanet archive
    """
    from PyAstronomy import pyasl
    import matplotlib.pylab as plt
    
    nexa = pyasl.NasaExoplanetArchive()
    
    # See what information is available
    cols = nexa.availableColumns()
    print()
    
    # Get all information for planet 'wasp-12 b'
    # By default, the search is case-insensitive
    print("Entry of Wasp-12 b")
    print(nexa.selectByPlanetName("Wasp-12 b"))
    
    print()
    # Get all data and plot ra vs. dec
    dat = nexa.getAllData()
    plt.plot(dat.ra, dat.dec, 'b.')
#     plt.show()


class SanityOfNasaExoplanetEU(unittest.TestCase):
  
  def setUp(self):
    pass
  
  def tearDown(self):
    pass
  
  def sanity_example(self):
    """
      Exoplanet EU example
    """
    from PyAstronomy import pyasl
    import matplotlib.pylab as plt
    
    eu = pyasl.ExoplanetEU()
    
    # See what information is available
    cols = eu.availableColumns()
    print(cols)
    
    print()
    # Get all data and plot planet Mass vs.
    # semi-major axis in log-log plot
    dat = eu.getAllData()
    plt.xlabel("Planet Mass [RJ]")
    plt.ylabel("Semi-major axis [AU]")
    plt.loglog(dat.plMass, dat.sma, 'b.')
#     plt.show()


  def sanity_exampleExoplanetEU2(self):
    """
      Example of ExoplanetEU2
    """
    from PyAstronomy import pyasl
    import matplotlib.pylab as plt
    
    # Instantiate exoplanetEU2 object
    v = pyasl.ExoplanetEU2()
    
    # Show the available data
    v.showAvailableData()
    print()
    
    # Get a list of all available column names
    acs = v.getColnames()
    print("Available column names: " + ", ".join(acs))
    print()
    
    # Select data by planet name (returns a dictionary)
    print(v.selectByPlanetName("CoRoT-2 b"))
    print()
    
    # Get all data as an astropy table
    at = v.getAllDataAPT()
    
    # Export all data as a pandas DataFrame
    pd = v.getAllDataPandas()
    
    # Plot mass vs. SMA
    plt.title("Mass vs. SMA")
    plt.xlabel("[" + v.getUnitOf("mass") + "]")
    plt.ylabel("[" + v.getUnitOf("semi_major_axis") + "]")
    plt.loglog(at["mass"], at["semi_major_axis"], 'b.')
#     plt.show()
    

## Defunct
# class SanityOfNasaExoplanetsOrg(unittest.TestCase):
#
#   def setUp(self):
#     pass
#
#   def tearDown(self):
#     pass
#
#   def sanity_example(self):
#     """
#       Exoplanets ORG example
#     """
#     from PyAstronomy import pyasl
#
#     # Instantiate the access class
#     epl = pyasl.ExoplanetsOrg()
#
#     # Show the available columns
#     epl.availableColumns()
#
#     # Get information in Kepler-5 b
#     d = epl.selectByPlanetName("kepler-5 b")
#
#     # Print whatever information has been received
#     print()
#     print("Information on Kepler-5 b")
#     print()
#     for k, v in list(d.items()):
#       print("%12s  %12s" % (k,str(v)))
#

      
class SanityOfFIP(unittest.TestCase):
  
  def setUp(self):
    pass
  
  def tearDown(self):
    pass
  
  def sanity_example(self):
    """
      Checking example of FIP.
    """
    from PyAstronomy import pyasl
    
    fip = pyasl.FirstIonizationPot()
    
    print("First ionization energy of Li = %4.2e +/- %4.2e eV" % fip.getFIP(3))
    print("First ionization energy of Protactinium = %4.2e +/- %4.2e eV" % fip.getFIP(91))
    
    # And the same using elemental symbols
    print()
    print("First ionization energy of Li = %4.2e +/- %4.2e eV" % fip.getFIP("Li"))
    print("First ionization energy of Protactinium = %4.2e +/- %4.2e eV" % fip.getFIP("Pa"))
    
    # Plot the first ionization energy as a function of atomic number
#     pyasl.plotFIP()


  def sanity_FIP(self):
    """
      Checking sanity of FIP.
    """
    from PyAstronomy import pyasl
  
    fip = pyasl.FirstIonizationPot()
    self.assertAlmostEqual(10.451260, fip.getFIP(53)[0], 6, msg="FIP for iodine does not match (an).")
    self.assertAlmostEqual(10.451260, fip.getFIP("I")[0], 6, msg="FIP for iodine does not match (sym).")
    
    self.assertAlmostEqual(25e-6, fip.getFIP(53)[1], 6, msg="FIP for iodine does not match (an).")
    self.assertAlmostEqual(25e-6, fip.getFIP("I")[1], 6, msg="FIP for iodine does not match (sym).")
    
    
class SanityOfKurucz(unittest.TestCase):
  
  def setUp(self):
    pass
  
  def tearDown(self):
    pass   
   
  def sanity_exmaple(self):
    """
      Sanity of Kurucz example
    """
    from PyAstronomy import pyasl
    
    km = pyasl.KuruczModels()
    # See what model grids are available
    print(km.availableGrids())
    
    # See whether model grid for log(metallicity) = 0.0
    # is available
    print(km.gridAvailable(0.0))
    
    # Obtain the model grid for solar metallicity
    mg = km.requestModelGrid(0.0)
    
    # See what Teffs and logg are available
    print("Teffs: ", mg.availableTeffs())
    print("Loggs: ", mg.availableLoggs())
    
    print()
    print()
    
    # Use simple access method to obtain a model.
    # The input is: Teff, logg, and log10(metallicity)
    model = pyasl.getKuruczModel(4250, 4.5, 0.1)



from __future__ import print_function, division
from PyAstronomy.pyaC import pyaPermanent as pp
from PyAstronomy import pyaC
import os
import gzip
import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE
import six


class Baraffe98Tracks:
    """
    Provide access to the evolutionary tracks of Baraffe et al. 98.

    Downloads the table (tab1-3.dat) pertaining to the paper
    Baraffe et al. 98, A&A 337, 403 and offers some helper
    method to access the individual tracks.
    """

    def _identifyModels(self):
        """
        Find all independent models in the data.

        Sets up the `models` attribute, which maps
        a model descriptor (metallicity, Y, L (mixing length),
        mass) to an array of indices, which specifies the
        fraction of the data containing that model.
        """
        self._models = {}
        metals = np.unique(self.dat[::, 0])
        ys = np.unique(self.dat[::, 1])
        ls = np.unique(self.dat[::, 2])
        masses = np.unique(self.dat[::, 3])
        nl = pyaC.NestedLoop([len(metals), len(ys), len(ls), len(masses)])
        for i in nl:
            indi = np.where(
                np.logical_and(
                    self.dat[::, 0] == metals[i[0]],
                    np.logical_and(
                        self.dat[::, 1] == ys[i[1]],
                        np.logical_and(
                            self.dat[::, 2] == ls[i[2]], self.dat[::, 3] == masses[i[3]]
                        ),
                    ),
                )
            )[0]
            if len(indi) > 0:
                self._models[
                    (metals[i[0]], ys[i[1]], ls[i[2]], masses[i[3]])
                ] = indi.copy()

    def __init__(self):
        pfs = pp.pyaFS.PyAFS()
        dfn = os.path.join("pyasl", "resBased", "baraffe98tracks.dat.gz")
        pfs.downloadToFile(
            "ftp://cdsarc.u-strasbg.fr/pub/cats/J/A+A/337/403/tab1-3.dat.gz",
            dfn,
            verbose=True,
            clobber=False,
        )
        self.dat = np.loadtxt(pfs.requestFile(dfn, openMethod=gzip.open))
        self._identifyModels()

        self.colToName = {
            0: "Met",
            1: "Y",
            2: "Lmix",
            3: "Mass",
            4: "Age",
            5: "Teff",
            6: "logg",
            7: "Mbol",
            8: "MV",
            9: "MR",
            10: "MI",
            11: "MJ",
            12: "MH",
            13: "MK",
        }
        self.nameToCol = {}
        for k, v in six.iteritems(self.colToName):
            self.nameToCol[v] = k

    def getUniqueValues(self, colName):
        """
        Find unique values in a column.

        Parameters
        ----------
        colName : string, {"Met","Y","Lmix","Mass","Age","Teff","logg","Mbol","MV","MR","MI","MJ","MH","MK"}
            The name of the column.

        Returns
        -------
        Unique values : array
            All unique values in that column.
        """
        if not colName in six.iterkeys(self.nameToCol):
            raise (
                PE.PyAValError(
                    "No column named '" + str(colName) + "'.",
                    solution="Choose one of: " + str(self.nameToCol.keys()),
                )
            )

        return np.unique(self.dat[::, self.nameToCol[colName]])

    def getModelData(self, model, columns=None):
        """
        Find data for a specific model.

        Parameters
        ----------
        model : tuple of float
            Specifies the model as a tuple containing four
            entries, which define: metallicity, initial
            helium mass fraction (Y), initial mixing length
            parameter (Lmix), and mass.
        columns : list of strings, optional
            A list of column names, which shall be included
            in the result. The default is including all
            columns.

        Returns
        -------
        Model : recarray
            The data pertaining to that particular model
            as a numpy recarray.
        """
        if not model in self._models:
            raise (
                PE.PyAValError(
                    "No model with parameters: " + str(model),
                    solution="Use, e.g., `getAvailableValues` to check what is available.",
                )
            )
        if columns is None:
            columns = list(self.nameToCol.keys())
        # Check validity of column names
        for col in columns:
            if not col in self.nameToCol.keys():
                raise (
                    PE.PyAValError(
                        "No column named '" + str(col) + "'.",
                        solution="Choose one of: " + str(self.nameToCol.keys()),
                        where="getModel",
                    )
                )
        # There is a valid model, return it as a recarray
        dt = []
        for col in columns:
            dt.append((col, float))
        result = np.recarray(shape=(len(self._models[model]),), dtype=dt)
        for col in columns:
            result[col] = self.dat[self._models[model], self.nameToCol[col]]
        return result

    def findModels(self, Met=None, Y=None, Lmix=None, Mass=None):
        """
        Find models with given parameters.

        Parameters
        ----------
        Met : float, optional
            The metallicity.
        Y : float, optional
            The initial helium mass fraction.
        Lmix : float, optional
            The initial mixing length parameter.
        Mass : float, optional
            The mass [solar masses]

        Returns
        -------
        Models : list of tuples
            A list of models specified as tuples
            defining: metallicity, Y, Lmix, and mass.
        """
        result = []
        for m in self._models:
            if Met is not None:
                if Met != m[0]:
                    continue
            if Y is not None:
                if Y != m[1]:
                    continue
            if Lmix is not None:
                if Lmix != m[2]:
                    continue
            if Mass is not None:
                if Mass != m[3]:
                    continue
            result.append(m)
        return result



"""
      subroutine occultnl(rl,c1,c2,c3,c4,b0,mulimb0,mulimbf,nb)
C Please cite Mandel & Agol (2002) if making use of this routine.
      implicit none
      integer i,j,nb,nr,i1,i2,nmax
      parameter (nmax=513)
      real*8 mulimbf(nb,5),pi,c1,c2,c3,c4,rl,bt0(nb),b0(nb),mulimb0(nb),
     &       mulimb(nb),mulimbp(nb),dt,t(nmax),th(nmax),r(nmax),sig,
     &       mulimb1(nb),mulimbhalf(nb),mulimb3half(nb),mulimb2(nb),
     &       sig1,sig2,omega,dmumax,fac,mu(nb),f1,f2
      pi=acos(-1.d0)
C  This routine uses the results for a uniform source to
C  compute the lightcurve for a limb-darkened source
C  (5-1-02 notes)
C Input:
C   rl        radius of the lens   in units of the source radius
C   c1-c4     limb-darkening coefficients
C   b0        impact parameter normalized to source radius
C Output:
C  mulimb0 limb-darkened magnification
C  mulimbf lightcurves for each component
C  
C  First, make grid in radius:
C  Call magnification of uniform source:
      call occultuniform(b0,rl,mulimb0,nb)
      i1=nb
      i2=1
      fac=0.d0
      do i=1,nb
        bt0(i)=b0(i)
        mulimbf(i,1)=1.d0
        mulimbf(i,2)=0.8d0
        mulimbf(i,3)=2.d0/3.d0
        mulimbf(i,4)=4.d0/7.d0
        mulimbf(i,5)=0.5d0
        mulimb(i)=mulimb0(i)
        if(mulimb0(i).ne.1.d0) then
          i1=min(i1,i)
          i2=max(i2,i)
        endif
        fac=max(fac,abs(mulimb0(i)-1.d0))
      enddo
C print,rl
      omega=4.*((1.d0-c1-c2-c3-c4)/4.+c1/5.+c2/6.+c3/7.+c4/8.)
      nr=2
      dmumax=1.d0
c      write(6,*) 'i1,i2 ',i1,i2
      do while (dmumax.gt.fac*1.d-3)
        do i=i1,i2
          mulimbp(i)=mulimb(i)
        enddo
        nr=nr*2
c        write(6,*) 'nr ',nr
        dt=0.5d0*pi/dble(nr)
        do j=1,nr+1
          t(j) =dt*dble(j-1)
          th(j)=t(j)+0.5d0*dt
          r(j)=sin(t(j))
        enddo
        sig=sqrt(cos(th(nr)))
        do i=i1,i2
          mulimbhalf(i) =sig**3*mulimb0(i)/(1.d0-r(nr))
          mulimb1(i)    =sig**4*mulimb0(i)/(1.d0-r(nr))
          mulimb3half(i)=sig**5*mulimb0(i)/(1.d0-r(nr))
          mulimb2(i)    =sig**6*mulimb0(i)/(1.d0-r(nr))
        enddo
        do j=2,nr
          do i=1,nb
            b0(i)=bt0(i)/r(j)
          enddo
C  Calculate uniform magnification at intermediate radii:
          call occultuniform(b0,rl/r(j),mu,nb)
C  Equation (29):
          sig1=sqrt(cos(th(j-1)))
          sig2=sqrt(cos(th(j)))
          dmumax=0.d0
          do i=i1,i2
            f1=r(j)*r(j)*mu(i)/(r(j)-r(j-1))
            f2=r(j)*r(j)*mu(i)/(r(j+1)-r(j))
            mulimbhalf(i) =mulimbhalf(i) +f1*sig1**3-f2*sig2**3
            mulimb1(i)    =mulimb1(i)    +f1*sig1**4-f2*sig2**4
            mulimb3half(i)=mulimb3half(i)+f1*sig1**5-f2*sig2**5
            mulimb2(i)    =mulimb2(i)    +f1*sig1**6-f2*sig2**6
            mulimb(i)=((1.d0-c1-c2-c3-c4)*mulimb0(i)+c1*mulimbhalf(i)*dt
     &        +c2*mulimb1(i)*dt+c3*mulimb3half(i)*dt+c4*mulimb2(i)*dt)
     &        /omega
            if(mulimb(i)+mulimbp(i).ne.0.d0) then 
              dmumax=max(dmumax,abs(mulimb(i)-mulimbp(i))/(mulimb(i)+
     &               mulimbp(i)))
            endif
          enddo
        enddo
      enddo
      do i=i1,i2
        mulimbf(i,1)=mulimb0(i)
        mulimbf(i,2)=mulimbhalf(i)*dt
        mulimbf(i,3)=mulimb1(i)*dt
        mulimbf(i,4)=mulimb3half(i)*dt
        mulimbf(i,5)=mulimb2(i)*dt
        mulimb0(i)=mulimb(i)
      enddo
      do i=1,nb
        b0(i)=bt0(i)
      enddo
      return
      end
      
      
      subroutine occultuniform(b0,w,muo1,nb)
      implicit none
      integer i,nb
      real*8 muo1(nb),w,b0(nb),z,pi,lambdae,kap0,kap1
      if(abs(w-0.5d0).lt.1.d-3) w=0.5d0
      pi=acos(-1.d0)
C  This routine computes the lightcurve for occultation
C  of a uniform source without microlensing  (Mandel & Agol 2002).
C Input:
C 
C  rs   radius of the source (set to unity)
C  b0   impact parameter in units of rs
C  w    occulting star size in units of rs
C 
C Output:
C  muo1 fraction of flux at each b0 for a uniform source
C 
C  Now, compute pure occultation curve:
      do i=1,nb
C  substitute z=b0(i) to shorten expressions
        z=b0(i)
C  the source is unocculted:
C  Table 3, I.
        if(z.ge.1.d0+w) then
          muo1(i)=1.d0
          goto 1
        endif
C  the  source is completely occulted:
C  Table 3, II.
        if(w.ge.1.d0.and.z.le.w-1.d0) then
          muo1(i)=0.d0
          goto 1
        endif
C  the source is partly occulted and the occulting object crosses the limb:
C  Equation (26):
        if(z.ge.abs(1.d0-w).and.z.le.1.d0+w) then
          kap1=acos(min((1.d0-w*w+z*z)/2.d0/z,1.d0))
          kap0=acos(min((w*w+z*z-1.d0)/2.d0/w/z,1.d0))
          lambdae=w*w*kap0+kap1
          lambdae=(lambdae-0.5d0*sqrt(max(4.d0*z*z-(1.d0+z*z-w*w)**2,
     &            0.d0)))/pi
          muo1(i)=1.d0-lambdae
        endif
C  the occulting object transits the source star (but doesn't
C  completely cover it):
        if(z.le.1.d0-w) muo1(i)=1.d0-w*w
 1      continue
      enddo
C muo1=1.d0-lambdae
      return
      end
"""

import numpy as np
from numpy import sqrt

class OccultNlPy:
    
    nb = 513
    
    def occultnl(self,rl,c1,c2,c3,c4,b0,mulimb0,mulimbf,nb):
# C Please cite Mandel & Agol (2002) if making use of this routine.
        pi = np.pi
# C  This routine uses the results for a uniform source to
# C  compute the lightcurve for a limb-darkened source
# C  (5-1-02 notes)
# C Input:
# C   rl        radius of the lens   in units of the source radius
# C   c1-c4     limb-darkening coefficients
# C   b0        impact parameter normalized to source radius
# C Output:
# C  mulimb0 limb-darkened magnification
# C  mulimbf lightcurves for each component
# C  
# C  First, make grid in radius:
# C  Call magnification of uniform source:
        mulimb0 = self.occultuniform(b0,rl,nb)
        # Account for one- vs zero-based indexing
        i1 = nb-1
        i2 = 1-1
        fac = 0.0
        bt0 = b0.copy()
        mulimbf = np.zeros((nb, 5))
        mulimb = mulimb0.copy()
        for i in range(nb):
            mulimbf[i,0] = 1.0
            mulimbf[i,1] = 0.8
            mulimbf[i,2] = 2.0/3.0
            mulimbf[i,3] = 4.0/7.0
            mulimbf[i,4] = 0.5
            if mulimb0[i] != 1.0:
                i1 = min(i1,i)
                i2 = max(i2,i)
            fac = max(fac,abs(mulimb0[i]-1.0))
# C print,rl
        omega = 4.*((1.0-c1-c2-c3-c4)/4.+c1/5.+c2/6.+c3/7.+c4/8.)
        nr = 2
        dmumax = 1.0
# c      write(6,*) 'i1,i2 ',i1,i2
        while True: #do while (dmumax.gt.fac*1.d-3)
            for i in range(i1,i2): #do i=i1,i2
                mulimbp[i] = mulimb[i]
        
            nr = nr*2
# c        write(6,*) 'nr ',nr
            dt = 0.5*pi/nr
        do j=1,nr+1
          t(j) =dt*dble(j-1)
          th(j)=t(j)+0.5d0*dt
          r(j)=sin(t(j))
        enddo
        sig=sqrt(cos(th(nr)))
        do i=i1,i2
          mulimbhalf(i) =sig**3*mulimb0(i)/(1.d0-r(nr))
          mulimb1(i)    =sig**4*mulimb0(i)/(1.d0-r(nr))
          mulimb3half(i)=sig**5*mulimb0(i)/(1.d0-r(nr))
          mulimb2(i)    =sig**6*mulimb0(i)/(1.d0-r(nr))
        enddo
        do j=2,nr
          do i=1,nb
            b0(i)=bt0(i)/r(j)
          enddo
C  Calculate uniform magnification at intermediate radii:
          call occultuniform(b0,rl/r(j),mu,nb)
C  Equation (29):
          sig1=sqrt(cos(th(j-1)))
          sig2=sqrt(cos(th(j)))
          dmumax=0.d0
          do i=i1,i2
            f1=r(j)*r(j)*mu(i)/(r(j)-r(j-1))
            f2=r(j)*r(j)*mu(i)/(r(j+1)-r(j))
            mulimbhalf(i) =mulimbhalf(i) +f1*sig1**3-f2*sig2**3
            mulimb1(i)    =mulimb1(i)    +f1*sig1**4-f2*sig2**4
            mulimb3half(i)=mulimb3half(i)+f1*sig1**5-f2*sig2**5
            mulimb2(i)    =mulimb2(i)    +f1*sig1**6-f2*sig2**6
            mulimb(i)=((1.d0-c1-c2-c3-c4)*mulimb0(i)+c1*mulimbhalf(i)*dt
     &        +c2*mulimb1(i)*dt+c3*mulimb3half(i)*dt+c4*mulimb2(i)*dt)
     &        /omega
            if(mulimb(i)+mulimbp(i).ne.0.d0) then 
              dmumax=max(dmumax,abs(mulimb(i)-mulimbp(i))/(mulimb(i)+
     &               mulimbp(i)))
            endif
          enddo
        enddo
      enddo
      do i=i1,i2
        mulimbf(i,1)=mulimb0(i)
        mulimbf(i,2)=mulimbhalf(i)*dt
        mulimbf(i,3)=mulimb1(i)*dt
        mulimbf(i,4)=mulimb3half(i)*dt
        mulimbf(i,5)=mulimb2(i)*dt
        mulimb0(i)=mulimb(i)
      enddo
      do i=1,nb
        b0(i)=bt0(i)
      enddo
      return
      end
    
    
    def occultuniform(self,b0,w,nb):

        if abs(w-0.5) < 1e-3:
            w = 0.5
        pi = np.pi
        
        muo1 = np.zeros(nb)
# C  This routine computes the lightcurve for occultation
# C  of a uniform source without microlensing  (Mandel & Agol 2002).
# C Input:
# C 
# C  rs   radius of the source (set to unity)
# C  b0   impact parameter in units of rs
# C  w    occulting star size in units of rs
# C 
# C Output:
# C  muo1 fraction of flux at each b0 for a uniform source
# C 
# C  Now, compute pure occultation curve:
        for i in range(nb):
# C  substitute z=b0(i) to shorten expressions
            z = b0[i]
# C  the source is unocculted:
# C  Table 3, I.
            if z >= 1+w:
                muo1[i] = 1
                continue
# C  the  source is completely occulted:
# C  Table 3, II.
            if (w >= 1) and (z <= w-1):
                muo1[i] = 0
                continue
# C  the source is partly occulted and the occulting object crosses the limb:
# C  Equation (26):
            if (z >= abs(1-w)) and (z <= 1+w):
                kap1 = np.arccos(min((1-w*w+z*z)/2/z,1))
                kap0 = np.arccos(min((w*w+z*z-1)/2/w/z,1))
                lambdae = w*w*kap0+kap1
                lambdae = (lambdae-0.5*sqrt(max(4*z*z-(1+z*z-w*w)**2,0)))/pi
                muo1[i] = 1-lambdae
# C  the occulting object transits the source star (but doesn't
# C  completely cover it):
            if (z <= 1-w):
                muo1[i] = 1-w*w
        return muo1



from __future__ import division
import re
import os.path
import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE
import six.moves as smo

class _FDs:
  
  def __init__(self, fn):
    self.fds = {}
    with open(fn) as f:
      for line in f:
        r = re.match(r"^\s*(\d+)\s+(\d+)\s+(.*)$", line)
        if r is None:
          continue
        derivative = int(r.group(1))
        errOrd = int(r.group(2))
        # Extract finite differences
        fdsStrings = r.group(3).split()
        coeffs = np.zeros(len(fdsStrings))
        offsets = np.arange(len(fdsStrings), dtype=int) - len(fdsStrings)//2
        for i, s in enumerate(fdsStrings):
          r = re.match(r"(-)?(\d+)(/(\d+))?", s)
          coeffs[i] = float(r.group(2))
          if r.group(4) is not None:
            coeffs[i] /= float(r.group(4))
          if r.group(1) is not None:
            coeffs[i] *= -1.0
        self.fds[(derivative, errOrd)] = (offsets, coeffs)
  
  def getFD(self, derivative, errOrd):
    """
    """
    if not (derivative, errOrd) in self.fds:
      raise(PE.PyAValError("No finite differences for derivative " + str(derivative) + \
                           ", error order " + str(errOrd), \
                           where="getFD", \
                           solution="Check derivative and error order."))
    return self.fds[(derivative, errOrd)]
          
          
         

def diffCFD(x, y, derivative, errOrd=2):
  """
    Calculate the numerical derivative using central finite distances.
    
    Parameters
    ----------
    x : array
        The abscissa values.
    y : array
        The ordinate values.
    derivative : int
        The derivative to calculate. Use, e.g., 1 to obtain
        the first derivative.
    errOrd : int
        The order of the error term. The default is 2. Other
        possibilities are 4, 6, and 8. Note the finite
        differences may not be available for all combinations.
    
    Returns
    -------
    indices : array of int
        The indices with respect to the input abscissa to which
        the derivative applies.
    der : array
        The derivative.
  """
  cfd = _FDs(os.path.join(os.path.dirname(__file__), "centralFiniteDiff.dat"))
  offs, fds = cfd.getFD(derivative, errOrd)
  if len(x) < len(offs):
    raise(PE.PyAValError("The length of the input is insufficient.", \
                         where="diffCFD"))
  h = (x[1]-x[0])
  if np.any(np.abs((x[1:]-x[0:-1])/(x[1]-x[0]) - 1.0) > 1e-6):
    raise(PE.PyAValError("The abscissa needs to be equally spaced.", \
                         where="diffCFD"))
  
  imin = abs(min(offs))
  imax = len(x) - imin
  der = np.zeros(imax-imin)
  for i in smo.range(imin, imax):
    der[i-imin] = np.sum(y[i+offs] * fds) / h**derivative
  
  return list(range(imin, imax)), der
  
  
  
  



from __future__ import print_function, division
import unittest
from PyAstronomy import funcFit as fuf
import os
import numpy as np


class TestFuncFitSanity(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        if os.path.isfile("saveState.tmp"):
            os.remove("saveState.tmp")

    def testsanity_saveState(self):
        gf = fuf.GaussFit1d()
        gf.assignValue({"A": 1, "mu": 2, "off": 3, "sig": 4, "lin": 5.0})
        gf.setRestriction({"A": [0, 10], "off": [None, 7.8]})
        gf.thaw(["A", "mu", "lin"])
        dat = gf.saveState("saveState.tmp", clobber=True)

        gf2 = fuf.GaussFit1d()
        gf2.restoreState("saveState.tmp")
        self.assertEqual(gf2.parameters(), gf.parameters())
        self.assertEqual(gf2.getRestrictions(), gf.getRestrictions())
        self.assertEqual(gf2.frozenParameters(), gf.frozenParameters())

        gf2.restoreState(dat)
        self.assertEqual(gf2.parameters(), gf.parameters())
        self.assertEqual(gf2.getRestrictions(), gf.getRestrictions())
        self.assertEqual(gf2.frozenParameters(), gf.frozenParameters())

    def testsanity_coordinateGrid(self):
        x = np.arange(5)
        y = np.arange(10) + 10.0
        z = np.arange(2) + 100.0
        g = fuf.coordinateGrid(x, y, z)
        self.assertEqual(g[0, 0, 0, 0], 0.0)
        self.assertEqual(g[0, 0, 0, 1], 10.0)
        self.assertEqual(g[0, 0, 0, 2], 100.0)
        self.assertEqual(g[1, 0, 1, 0], 1.0)
        self.assertEqual(g[1, 2, 1, 1], 12.0)
        self.assertEqual(g[1, 2, 1, 2], 101.0)

    def testsanity_fminArgs(self):
        r"""
        Test handling of fminArgs
        """
        np.random.seed(101)

        fwhm = 10833/100_000
        snr = 100

        # Wvl and 'data' (independent, without signal)
        w = np.arange(10830, 10836, fwhm)
        d = np.random.normal(1, 1/snr, len(w))
        e = np.ones_like(w)/snr

        gf = fuf.GaussFit1d()
        gf["mu"] = 10833
        gf["off"] = 1.0
        gf["A"] = -0.1
        # Assumption
        gf["sig"] = fwhm*3

        # 'Real' estimate from the data
        gf.thaw(["A"])
        gf.fit(w, d, yerr=e, minAlgo='spfmp', fminArgs={'disp': 0})



from PyAstronomy.pyaC import pyaErrors as PE
from datetime import datetime as dt

def decimalYear(date):
    """
    Calculates the decimal representation of a date, e.g., 2013.12.
    
    The code uses Python's datetime package to determine the fractional
    year. Thus, leap years are taken into account. There may still be
    issues with time zones or daylight saving times etc..
    
    Code from:
    http://stackoverflow.com/questions/6451655/python-how-to-convert-datetime-dates-to-decimal-years

    Parameters
    ----------
    date : python date instance
        The input date (and time).
      
    Returns
    -------
    Decimal year : float
        Decimal representation of the date.
    """
    def s(date):
      # returns seconds since epoch
      return (date - dt(1900,1,1)).total_seconds()

    year = date.year
    startOfThisYear = dt(year=year, month=1, day=1)
    startOfNextYear = dt(year=year+1, month=1, day=1)

    yearElapsed = s(date) - s(startOfThisYear)
    yearDuration = s(startOfNextYear) - s(startOfThisYear)
    fraction = yearElapsed/yearDuration

    return date.year + fraction


def decimalYearGregorianDate(date, form="datetime"):
  """
    Convert decimal year into gregorian date.
    
    Formally, the precision of the result is one microsecond.
    
    Parameters
    ----------
    date : float
        The input date (and time).
    form : str, optional
        Output format for the date. Either one of the following strings
        defining a format: "dd-mm-yyyy [hh:mm:ss]", "yyyy-mm-dd [hh:mm:ss]",
        where the term in braces is optional, or 
        "tuple" or "datetime". If 'tuple' is specified, the result will be a
        tuple holding (year, month, day, hour, minute, second, microseconds).
        In the case of "datetime" (default), the result will be a
        datetime object.
      
    Returns
    -------
    Gregorian date : str, tuple, or datetime instance
        The gregorian representation of the input in the specified format.
        In case of an invalid format, None is returned.
  """
  def s(date):
    # returns seconds since epoch
    return (date - dt(1900,1,1)).total_seconds()
  
  # Shift the input of 1e-2 microseconds
  # This step accounts for rounding issues.
  date += 1e-5/(365.*86400.0)
  
  year = int(date)
  yearFraction = float(date) - int(date)
  startOfThisYear = dt(year=year, month=1, day=1)
  startOfNextYear = dt(year=year+1, month=1, day=1)
  secondsInYear = (s(startOfNextYear) - s(startOfThisYear)  ) * yearFraction
  # Find the month
  m = 1
  while m<=12 and s(dt(year=year, month=m, day=1)) - s(startOfThisYear) <= secondsInYear: m+=1
  m-=1
  # Find the day
  d = 1
  tdt = dt(year=year, month=m, day=d)
  while s(tdt) - s(startOfThisYear) <= secondsInYear:
    d+=1
    try: tdt=dt(year=year, month=m, day=d)
    except: break
  d-=1 
  # Find the time
  secondsRemaining = secondsInYear + s(startOfThisYear) - s(dt(year=year, month=m, day=d))
  hh = int(secondsRemaining/3600.)
  mm = int((secondsRemaining - hh*3600) / 60.)
  ss = int(secondsRemaining - hh*3600 - mm * 60) 
  ff = secondsRemaining - hh*3600 - mm * 60 - ss
  
  # Output formating
  if "tuple" == form:
    r = (year, m, d, hh, mm, ss, int(ff*1000))
  elif "datetime" == form:
    r = dt(year, m, d, hh, mm, ss, int(ff*1000))
  elif "dd-mm-yyyy" in form:
    r = str("%02i-%02i-%04i" % (d,m,year))
    if "hh:mm:ss" in form:
      r+=str(" %02i:%02i:%02i" % (hh,mm,ss))
  elif "yyyy-mm-dd" in form:
    r = str("%04i-%02i-%02i" % (year,m,d))
    if "hh:mm:ss" in form:
      r+=str(" %02i:%02i:%02i" % (hh,mm,ss))
  else:
    raise(PE.PyAValError("Invalid input form of `form` parameter.", \
                         where="gregorianDate"))
    return None
  return r  


# -*- coding: utf-8 -*-
import os
import re
import argparse
import configparser
import hashlib
import datetime
import autopep8


class CandidateBlock:

    minLength = 10

    def __init__(self):
        self.lines = []
        # The starting line (in the file) of the example
        self.startLine = None
        # True if the block is long enough to be an example
        self.lengthCheck = False
        # True if import statement has been found
        self.hasImport = False
        # Level if indention
        self.indentLevel = 1e100
        # Is this thought to be an example
        self.isExample = False
        # Is manually specified as example (by ".. IsPyAExample" directive)
        self.hasExampleDirective = False

    def characterize(self):
        """
        Check whether this block is a potential example.
        """
        # Check block length (count only nonempty lines)
        no = 0
        for l in self.lines:
            r = re.match("(\s*)[^\s]+.*", l)
            if r is not None:
                no += 1
                if len(r.group(1)) < self.indentLevel:
                    self.indentLevel = len(r.group(1))
            if re.match("\s*import\s.*", l) is not None:
                self.hasImport = True
        if no > self.minLength:
            self.lengthCheck = True

        if (self.lengthCheck and self.hasImport) or self.hasExampleDirective:
            self.isExample = True

    def commentShow(self):
        for i in range(len(self.lines)):
            r = re.match("(\s*)([^#]*\.show().*)", self.lines[i])
            if r is not None:
                self.lines[i] = r.group(1) + "# " + r.group(2)
                return

    def format(self, indentLevel=4, commentShow=True):
        if commentShow:
            self.commentShow()
        result = []
        indentor = "".join([" " for i in range(indentLevel)])
        result.append(indentor + "# Automatically extracted example:\n")
        result.append(indentor + "# Starting line: " +
                      str(self.startLine) + "\n")
        for l in self.lines:
            r = re.match("\s{" + str(self.indentLevel) + "}(.*)", l)
            if r is not None:
                result.append(indentor + r.group(1) + "\n")
        for i, l in enumerate(result):
            if l.find("from __future__ import") != -1:
                result[i] = "# " + result[i]
                break
        for i, l in enumerate(result):
            if l.find("import matplotlib.pylab as plt") != -1:
                result.insert(i+1, ((len(l) - len(l.lstrip())) * " ") + "plt.show = lambda : None\n" )
            if l.find("from PyAstronomy import pyaGui") != -1:
                result.insert(i+1, ((len(l) - len(l.lstrip())) * " ") + "plt.show = lambda : None\n" )
        return result

    def executeBlock(self, fn="tmp.py"):
        self.commentShow()
        f = open(fn, 'w')
        for l in self.lines:
            r = re.match("\s{" + str(self.indentLevel) + "}(.*)", l)
            if r is not None:
                f.write(r.group(1) + "\n")
        f.close()
        try:
            exec(open(fn), {})
        except:
            pass


class FileBlocks:

    def _findBlocks(self, lines):
        """
        Find potential example blocks and save them into self.blocks
        as CandidateBlock objects.
        """
        isInBlock = False
        indent = 0
        lineNo = 0
        lastLine_IsPyaExample = False

        for l in lines:
            lineNo += 1
            r = re.match(".. IsPyAExample", l)
            if r is not None:
                lastLine_IsPyaExample = True
                continue
            r = re.match("(\s*)::\s*", l)
            if r is not None:
                isInBlock = True
                block = CandidateBlock()
                # Classifies as example if specified manually
                block.hasExampleDirective = lastLine_IsPyaExample
                block.startLine = lineNo
                indent = len(r.group(1))
                lastLine_IsPyaExample = False
                continue
            r = re.match("(\s*)([^\s]+)", l)
            if r is not None:
                # Not empty line
                lastLine_IsPyaExample = False
            if not isInBlock:
                continue
            if r is None:
                # Empty line
                block.lines.append(l)
                continue
            if len(r.group(1)) == indent:
                isInBlock = False
                self.blocks.append(block)
                continue
            block.lines.append(l)
        if isInBlock:
            # File seems to end with a potential example block
            self.blocks.append(block)

    def _characterizeBlocks(self):
        for i in range(len(self.blocks)):
            self.blocks[i].characterize()

    def noOfExamples(self):
        i = 0
        for b in self.blocks:
            if b.isExample:
                i += 1
        return i

    def summary(self):
        """
        Print a summary of the analysis
        """
        print("Analyzed file: ", self.fn)
        print("Found " + str(len(self.blocks)) + " potential example blocks")
        i = self.noOfExamples()
        print("  of which " + str(i) + " have been characterized as example.")
        print()

    def executeExamples(self):
        for i in range(len(self.blocks)):
            if self.blocks[i].isExample:
                self.blocks[i].executeBlock()

    def __checkCreateTestDir(self):
        if os.path.isdir(self.snDir):
            return
        os.mkdir(self.snDir)

    def pep8blocks(self):
        """
        """
        ll = open(self.fn).readlines()
        
        add = 0
        for b in self.blocks:
            if not b.isExample:
                continue
            
            print("Startline: ", b.startLine)
            afteradd = 0
            lls = [l[b.indentLevel:] for l in b.lines]
            for i in range(len(lls)):
                lls[i] = lls[i].rstrip("\n") + "\n"
            print("PRE")
            for j, l in enumerate(lls):
                print(j, " '"+l+"'")
            print("END PRE")
            nl = autopep8.fix_code(''.join(lls), options={"max_line_length":1000})
            nls = nl.split("\n")[0:-1]
            nls = [" "*4+l.rstrip("\n") for l in nls]
            print("POST")
            for j, l in enumerate(nls):
                print(j, " '"+l.rstrip("\n")+"'")
            print("END POST")
            print("lens: ", len(nls), len(b.lines))
            if len(nls) > len(b.lines):
                print("Adding ", len(nls) - len(b.lines), " additional lines")
                for i in range(len(nls) - len(b.lines)):
                    ll.insert(b.startLine+1, "")
                    afteradd += 1

            for i in range(len(nls)):
                print("Old: '" + ll[i+b.startLine+add] + "'")
                print("New: '" + nls[i] + "'")
                ll[i+b.startLine+add] = nls[i] +"\n"
            add += afteradd
        
        for l in ll:            
            print(l, end="")
        open(self.fn, 'wt').writelines(ll)
            

    def toSanityCheckFile(self):
        """
        Write all blocks identified as examples to a file identified by
        the sha1 hash of the full file name.
        """
        if self.noOfExamples() == 0:
            return
        s1 = hashlib.sha1()
        s1.update(self.fn.encode("utf-8"))
        ofn = os.path.join(self.snDir, "sanity_" + s1.hexdigest() + ".py")
        f = open(ofn, 'w')
        f.write("# File written by TBD at " +
                str(datetime.datetime.now()) + "\n")
        f.write("# Examples extracted from file: " + self.fn + "\n\n\n")
        for i in range(len(self.blocks)):
            if self.blocks[i].isExample:
                f.write("def sanity_line_" +
                        str(self.blocks[i].startLine) + "():\n")
                lines = self.blocks[i].format()
                f.writelines(lines)
                f.write("\n\n")
        f.close()

    def __init__(self, fn, snDir="sanityies"):

        if not os.path.isfile(fn):
            print("No such file: ", fn)
            return

        self.fn = fn
        self.blocks = []
        self.snDir = snDir
        self.__checkCreateTestDir()

        lines = open(fn).readlines()
        self._findBlocks(lines)
        self._characterizeBlocks()


class Walker:

    def __init__(self, rootPath="../", fnRegex=".*\.rst"):
        self.rootPath = rootPath
        self.fnRegex = fnRegex

    def walk(self):
        for root, dirs, files in os.walk(self.rootPath):
            for fn in files:
                if re.match(self.fnRegex, fn) is not None:
                    yield os.path.join(root, fn)


parser = argparse.ArgumentParser(description='Options')
parser.add_argument('--cf', dest='cf', default="dysSetup.cfg",
                    help="Name of config file.")
parser.add_argument('--root', dest='root',
                    help="The root directory for the example search.")
parser.add_argument('--apep8', dest='apep8', action="store_true", 
                    help="Convert detected example to PEP8.")
args = parser.parse_args()


config = configparser.ConfigParser()
config.read(args.cf)

if args.root is None:
    rootPath = config.get("Walker", "rootPath")
else:
    rootPath = args.root

fnRegex = config.get("Walker", "fnRegex")
walker = Walker(rootPath=rootPath, fnRegex=fnRegex)

for fn in walker.walk():
    fbs = FileBlocks(fn)
    fbs.summary()
    if args.apep8:
        fbs.pep8blocks()
    fbs.toSanityCheckFile()



from PyAstronomy import pyaC
from numpy import arctan2, sin, cos, tan

def positionAngle(ra1, dec1, ra2, dec2, positive=True):
  """
    Compute the position angle.
  
    The position angle is measured from the first position
    from North through East. If the `positive` flag is set
    True (default) the result will be given as an angle from
    0 to 360 degrees. If the flag is set False, the scale is
    -180 to 180 degrees with negative number increasing from
    North through West; note that this is the behavior of
    the posAng IDL routine.
    
    Parameters
    ----------
    ra1 : float
        Right ascension of first object [deg].
    dec1 : float
        Declination of first object [deg].
    ra2 : float
        Right ascension of second object [deg].
    dec2 : float
        Declination of second object [deg].
    positive : boolean, optional
        If True (default), the output will be
        given as an angle between 0 and 360
        degrees. Otherwise, the angle ranges
        from -180 to +180 degrees.
    
    Returns
    -------
    Position angle : float
        The position angle in degrees.
    
  
    Notes
    -----
    
    .. note:: This function was ported from the IDL Astronomy User's Library.
    
    :IDL - Documentation:  
    
     NAME:
           POSANG
     PURPOSE:
           Computes rigorous position angle of source 2 relative to source 1
           
     EXPLANATION:
           Computes the rigorous position angle of source 2 (with given RA, Dec) 
           using source 1 (with given RA, Dec) as the center.
     
     CALLING SEQUENCE:
           POSANG, U, RA1, DC1, RA2, DC2, ANGLE
    
     INPUTS:
           U    -- Describes units of inputs and output:
                   0:  everything radians
                   1:  RAx in decimal hours, DCx in decimal
                           degrees, ANGLE in degrees
           RA1  -- Right ascension of point 1
           DC1  -- Declination of point 1
           RA2  -- Right ascension of point 2
           DC2  -- Declination of point 2
    
       OUTPUTS:
           ANGLE-- Angle of the great circle containing [ra2, dc2] from
                   the meridian containing [ra1, dc1], in the sense north
                   through east rotating about [ra1, dc1].  See U above 
                   for units.
    
       PROCEDURE:
           The "four-parts formula" from spherical trig (p. 12 of Smart's
           Spherical Astronomy or p. 12 of Green' Spherical Astronomy).
    
       EXAMPLE:
           For the star 56 Per, the Hipparcos catalog gives a position of 
           RA = 66.15593384, Dec = 33.94988843 for component A, and 
           RA = 66.15646079, Dec =  33.96100069 for component B.   What is the
           position angle of B relative to A?
    
           IDL> RA1 = 66.15593384/15.d   & DC1 = 33.95988843
           IDL> RA2 = 66.15646079/15.d   & DC2 = 33.96100069
           IDL> posang,1,ra1,dc1,ra2,dc2, ang
                will give the answer of ang = 21.4 degrees
       NOTES:
           (1) If RA1,DC1 are scalars, and RA2,DC2 are vectors, then ANGLE is a
           vector giving the position angle between each element of RA2,DC2 and 
           RA1,DC1.   Similarly, if RA1,DC1 are vectors, and RA2, DC2 are scalars,
           then DIS is a vector giving the position angle of each element of RA1, 
           DC1 and RA2, DC2.    If both RA1,DC1 and RA2,DC2 are vectors then ANGLE 
           is a vector giving the position angle between each element of RA1,DC1 
           and the corresponding element of RA2,DC2.    If then vectors are not the
           same length, then excess elements of the longer one will be ignored.
    
           (2) Note that POSANG is not commutative -- the position angle between
            A and B is theta, then the position angle between B and A is 180+theta 
       PROCEDURE CALLS:
            ISARRAY()
       HISTORY:
           Modified from GCIRC, R. S. Hill, RSTX, 1 Apr. 1998
           Use V6.0 notation W.L. Mar 2011
  """

  # Convert into rad               
  rarad1 = pyaC.degtorad(ra1)
  rarad2 = pyaC.degtorad(ra2)
  dcrad1 = pyaC.degtorad(dec1)
  dcrad2 = pyaC.degtorad(dec2)

  radif  = rarad2 - rarad1

  angle  = arctan2(sin(radif), cos(dcrad1)*tan(dcrad2)-sin(dcrad1)*cos(radif))

  result = pyaC.radtodeg(angle)
  
  if positive and (result < 0.0):
    result+= 360.0

  return result  



from __future__ import division
from . import pal
import PyAstronomy.funcFit as fuf
from PyAstronomy.pyaC import pyaErrors as PE
import six

class SecPalLC(fuf.OneDFit):
  """
    Calculate and fit analytical secondary transit light-curves using the formulae
    provided by Pal 2008.
    
    .. note :: The **evaluation of elliptical integrals** is essential in
               calculating the transit model. While both
               the *mpmath* module and the *Boost* libraries implement those
               integrals, it is the Boost library, which
               evaluates them by far more quickly. Yet, the support
               for Boost has to be added manually. 
    
    *Model parameters*:
      - `p` - Radius ratio between planet and star.
      - `a` - Semi-major axis of planetary orbit [stellar radii].
      - `i` - Inclination of orbit in degrees (90 deg is *edge on* view).
      - `T0` - Time offset of secondary transit center.
      - `per` - Period of planetary orbit.
      - `b` - Describes the flux ratio between a stellar companion and the main star (default is 0).
      - `brat` - Brightness ratio between planet and star (one mean that planet has the same brightness as the star).

    @TODO - subtractOne docu

    This class inherits the functionality of funcFit's OneDFit object.
    You can, therefore, set and get the parameter using the brackets:
      e.g., pallc["p"] = 0.12345

    .. warning::
        Time units have to be consistent.
  """
  
  def __init__(self, subtractOne=True):
    fuf.OneDFit.__init__(self, ["p", "a", "i", "T0", "per", "brat"  ,"b"])
    self.freeze(["p", "a", "i", "T0", "per", "b", "brat"])
    self.pal = pal.PalLC()
    self.pal["quadLimb"] = 0.0
    self.pal["linLimb"] = 0.0
    self.pal["b"] = 0.0
    self.subtractOne = subtractOne
    self.setRootName("SecPal")

  def evaluate(self, time):
    """ 
     Calculate a light curve for secondary transit.
        
     Parameters
     ----------
     time : array
         An array of time points at which the light curve
         shall be calculated.
        
     .. note:: time = 0 -> Planet is exactly in the line of sight (phase = 0).

     Returns
     -------
     Model : array
         The analytical light curve.
    """
    # Assign all variables
    for k in six.iterkeys(self.propMap):
      if k == "brat" or k == "b":
        continue
      self.pal[k] = self[self.propMap[k]]
    # Calculate LC
    df = ((1.0 - self.pal.evaluate(time)) / self["p"]**2 * self["brat"])
    # Account for possible stellar companion
    y = 1.0 - ( df / (1.0 + self["b"]) )
    if self.subtractOne:
      y -= 1.0
    return y



# -*- coding: utf-8 -*-
import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE


def localTime(utc, lon, diff=True):
  """
    Computes the Local Time for a given UTC at a given geographic longitude.
    
    The local time is computed as UTC + LONGITUDE/15.
    
    Parameters
    ----------
    utc : float or array
        The time in UTC in hours.
    lon : float or array
       The geographic (East) longitude in DEGREES for which
       local time should be calculated.
    diff : boolean, optional
         If True (default), returns the difference in HOURS between
         UTC and local time.

    Returns
    -------
    Time : float or array
        Local time in HOURS (0 - 24) for given geographic
        longitude and UTC.
    Time difference : float or array
        The difference between local and UTC time in
        hours (only returned if `diff` is True)
  """

  utc = np.array(utc, ndmin=1)
  lon = np.array(lon, ndmin=1)
  if lon.size != utc.size:
    raise(PE.PyAValError("You need to specify the same number of longitudes and times", \
          solution="Make `lon` and `utc` arrays have the same length.", \
          where="localTime"))
  indi = np.where(np.logical_or(lon<0.0, lon>360.))[0]
  if len(indi) > 0:
    raise(PE.PyAValError("Longitude needs to be in the range 0-360.", \
          solution="Change the input.", \
          where="localTime"))
    
  localtime = utc + lon/15.
  localtime = localtime % 24.0

  if diff == True:
    return localtime, lon/15.
  else:
    return localtime




from PyAstronomy.pyaC import ImportCheck

ic = ImportCheck(["six.moves.tkinter", "matplotlib"], required=["six.moves.tkinter", "matplotlib"])

from .pyaPicker import Picker
from .ffmodelExplorer import FFModelExplorer, FFModelExplorerList, FFModelPlotFit, ffmodelExplorer
from .continuumFinder import ContinuumInteractive
from .interactiveGV import IAGVFit


from __future__ import print_function, division
import numpy as np
import os
import gzip
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy.pyasl import _ic 
import six.moves as smo
from PyAstronomy.pyaC import pyaPermanent as pp
from astropy.table import Table as _atable


class MMSCETSTable(pp.PyAUpdateCycle):
  """
    Access to the "Modern Mean Stellar Color and Effective Temperature Sequence" table. 
    
    A compilation of data in the form of a table entiled `"A Modern Mean Stellar Color and Effective Temperature
    Sequence for O9V-Y0V Dwarf Stars" <http://www.pas.rochester.edu/~emamajek/EEM_dwarf_UBVIJHK_colors_Teff.txt>`_
    is made available by E. Mamajek.
    
    In particular, the following columns are provided:
    SpT, Teff, logT, BCv, Mv, logL, B-V, Bt-Vt, U-B, V-Rc, V-Ic, V-Ks, J-H, H-K, Ks-W1, Msun, logAge, b-y,
    #SpT, M_J, M_Ks, Mbol".
    
    Much of the material (but not all) is published in Pecaut & Mamajek (2013, ApJS, 208, 9), which should
    be cited when using the data, and another reference for a fraction of the data is
    Pecaut, Mamajek, & Bubar (2012, ApJ 756, 154). More information can be found on the E. Mamajek's
    website and within the data file itself (use the getContent method to get the entire file).
    
    Entries without a valid entry are represented by NaN in this class.
    
    .. note:: The author of the table, E. Mamajek, states "Please email me if you use the table in your research
              and/or have any questions. - EEM". 
  """
  
  def __init__(self, skipUpdate=False, forceUpdate=False):
    
    if not _ic.check["astropy"]:
      raise(PE.PyARequiredImport("The 'astropy' package is not installed. astropy is required to read VO tables.", \
                                 solution="Please install 'astropy'."))

    
    configFilename = os.path.join("pyasl", "pyasl", "mamatable.cfg")
    pp.PyAUpdateCycle.__init__(self, configFilename, "MamajekTableUpdate", updateCycle=30)
    self.dataFileName = os.path.join("pyasl", "pyasl", "mamatable.gz")
    self._fs = pp.PyAFS()
    if forceUpdate:
      self._update(self._download)
    elif (self.needsUpdate() or (not self._fs.fileExists(self.dataFileName))) and (not skipUpdate):
      # Download data if data file does not exist or
      # regular update is indicated
      self._update(self._download)
    self._readData()
  
  def getTable(self):
    """
      Get table as astropy table object.
    """
    return self._tdata.copy()

  def getContent(self):
    """
      Get content of data file.
      
      The file contains a lot more information about
      the sources of the data, a change log etc..
      
      Returns
      -------
      content : list of strings
          The content of the file
    """
    return self._lines[:]
  
  def availableColumns(self):
    """
      Returns a list of available column names.
    """
    return self._tdata.columns[:]
  
  def getColumn(self, col, asarray=True):
    """
      Get data of specific column.
      
      Parameter
      ---------
      col : string
          Column specifier
      asarray : boolean, optional
          If True (default), the data will be converted into
          a numpy array. Otherwise, an astropy Column object
          is returned.
          
      Returns
      -------
      column data : array or column object
          Data for the specified column.
    """
    if asarray:
      return np.array(self._tdata[col])
    return self._tdata[col]
  
  def _readData(self):
    """
      Read the data from the file
    """
    self._lines = self._fs.requestFile(self.dataFileName, 'rt', gzip.open).readlines()
    # Table lines
    headerFound = False
    headerLine = None
    for n, l in enumerate(self._lines):
      if l.startswith("#SpT") and (not headerFound):
        # The header line was found
        headerLine = n
        headerFound = True
      elif l.startswith("#SpT") and headerFound:
        # This is the line after the table ends
        break
    tablelines = self._lines[headerLine:n]
    # Replace dot-placeholders by NaN
    for i in smo.range(len(tablelines)):
      for j in smo.range(5,2,-1):
        tablelines[i] = tablelines[i].replace("."*j, "NaN")
    # Read data
    self._tdata = _atable.read(tablelines, format="ascii")
      
  def _download(self):
    """
      Download data.
    """   
    self._fs.downloadToFile("http://www.pas.rochester.edu/~emamajek/EEM_dwarf_UBVIJHK_colors_Teff.txt", self.dataFileName, clobber=True,
                            verbose=False, openMethod=gzip.open)
    


# -*- coding: utf-8 -*-
import numpy as np

def gal_uvw(ra, dec, pmra, pmdec, distance, vrad, lsr=None):
    """
    Calculate the Galactic space velocity (U,V,W)
    
    .. note::  In contrast to the IDL implementation, U is taken positive toward the
               galactic center.
    
    Parameters
    ----------
    ra, dec : float or array
        Position in degrees
    pmra, pmdec : float or array
        Proper motion in mas/yr
    distance : float or array
        Distance in pc
    vrad : float or array
        Radial velocity in km/s
    lsr : three-float tuple, optional
        If given, the output UVW velocities are corrected for the
        solar motion (U,V,W)_Sun to the local standard of rest (LSR).
        Use, e.g., lsr = (8.5, 13.38, 6.49) from Coșkunoǧlu et al. 2011,
        MNRAS 412, 1237-1245. Be aware that the value of the solar motion
        through the LSR remains poorly determined.

    Returns
    -------
    U, V, W : float or arrays
        U: Velocity (km/s) positive toward the Galactic center
        V: Velocity (km/s) positive in the direction of Galactic rotation
        W: Velocity (km/s) positive toward the North Galactic Pole
    
    
    Notes
    -----
    
    .. note:: This function was ported from the IDL Astronomy User's Library.
    :IDL - Documentation:
    
    NAME:
        GAL_UVW
    PURPOSE:
        Calculate the Galactic space velocity (U,V,W) of star
    EXPLANATION:
        Calculates the Galactic space velocity U, V, W of star given its
        (1) coordinates, (2) proper motion, (3) distance (or parallax), and
        (4) radial velocity.
    CALLING SEQUENCE:
        GAL_UVW [/LSR, RA=, DEC=, PMRA= ,PMDEC=, VRAD= , DISTANCE=
                    PLX= ]
    OUTPUT PARAMETERS:
            U - Velocity (km/s) positive toward the Galactic *anti*center
            V - Velocity (km/s) positive in the direction of Galactic rotation
            W - Velocity (km/s) positive toward the North Galactic Pole
    REQUIRED INPUT KEYWORDS:
            User must supply a position, proper motion,radial velocity and distance
            (or parallax).    Either scalars or vectors can be supplied.
        (1) Position:
            RA - Right Ascension in *Degrees*
            Dec - Declination in *Degrees*
        (2) Proper Motion
            PMRA = Proper motion in RA in arc units (typically milli-arcseconds/yr)
            PMDEC = Proper motion in Declination (typically mas/yr)
        (3) Radial Velocity
            VRAD = radial velocity in km/s
        (4) Distance or Parallax
            DISTANCE - distance in parsecs
                    or
            PLX - parallax with same distance units as proper motion measurements
                typically milliarcseconds (mas)

    OPTIONAL INPUT KEYWORD:
            /LSR - If this keyword is set, then the output velocities will be
                corrected for the solar motion (U,V,W)_Sun = (-8.5, 13.38, 6.49)
                (Coskunoglu et al. 2011 MNRAS) to the local standard of rest.
                Note that the value of the solar motion through the LSR remains
                poorly determined.
        EXAMPLE:
            (1) Compute the U,V,W coordinates for the halo star HD 6755.
                Use values from Hipparcos catalog, and correct to the LSR
            ra = ten(1,9,42.3)*15.    & dec = ten(61,32,49.5)
            pmra = 627.89  &  pmdec = 77.84         ;mas/yr
            dis = 144    &  vrad = -321.4
            gal_uvw,u,v,w,ra=ra,dec=dec,pmra=pmra,pmdec=pmdec,vrad=vrad,dis=dis,/lsr
                ===>  u=154  v = -493  w = 97        ;km/s

            (2) Use the Hipparcos Input and Output Catalog IDL databases (see
            http://idlastro.gsfc.nasa.gov/ftp/zdbase/) to obtain space velocities
            for all stars within 10 pc with radial velocities > 10 km/s

            dbopen,'hipparcos,hic'      ;Need Hipparcos output and input catalogs
            list = dbfind('plx>100,vrad>10')      ;Plx > 100 mas, Vrad > 10 km/s
            dbext,list,'pmra,pmdec,vrad,ra,dec,plx',pmra,pmdec,vrad,ra,dec,plx
            ra = ra*15.                 ;Need right ascension in degrees
            GAL_UVW,u,v,w,ra=ra,dec=dec,pmra=pmra,pmdec=pmdec,vrad=vrad,plx = plx
            forprint,u,v,w              ;Display results
    METHOD:
            Follows the general outline of Johnson & Soderblom (1987, AJ, 93,864)
            except that U is positive outward toward the Galactic *anti*center, and
            the J2000 transformation matrix to Galactic coordinates is taken from
            the introduction to the Hipparcos catalog.
    REVISION HISTORY:
            Written, W. Landsman                       December   2000
            fix the bug occuring if the input arrays are longer than 32767
            and update the Sun velocity           Sergey Koposov June 2008
            vectorization of the loop -- performance on large arrays
            is now 10 times higher                Sergey Koposov December 2008
    """

    cosd = np.cos(np.deg2rad(dec))
    sind = np.sin(np.deg2rad(dec))
    cosa = np.cos(np.deg2rad(ra))
    sina = np.sin(np.deg2rad(ra))

    # Equivalent of 1 AU/yr in km/s   
    k = 4.74047     
    a_g = np.array([[0.0548755604, +0.4941094279, -0.8676661490], \
                    [0.8734370902, -0.4448296300, -0.1980763734], \
                    [0.4838350155, 0.74698224450, +0.4559837762]])

    # Parallax [mas]
    plx = 1e3 / distance
    
    vec1 = vrad
    vec2 = k * pmra / plx
    vec3 = k * pmdec / plx

    u = (a_g[0,0] * cosa * cosd + a_g[1,0] * sina * cosd + a_g[2,0] * sind) * vec1 + (-a_g[0,0] * sina + a_g[1,0] * cosa) * vec2 + (-a_g[0,0] * cosa * sind - a_g[1,0] * sina * sind + a_g[2,0] * cosd) * vec3
    v = (a_g[0,1] * cosa * cosd + a_g[1,1] * sina * cosd + a_g[2,1] * sind) * vec1 + (-a_g[0,1] * sina + a_g[1,1] * cosa) * vec2 + (-a_g[0,1] * cosa * sind - a_g[1,1] * sina * sind + a_g[2,1] * cosd) * vec3
    w = (a_g[0,2] * cosa * cosd + a_g[1,2] * sina * cosd + a_g[2,2] * sind) * vec1 + (-a_g[0,2] * sina + a_g[1,2] * cosa) * vec2 + (-a_g[0,2] * cosa * sind - a_g[1,2] * sina * sind + a_g[2,2] * cosd) * vec3
    # Change sign convention for U (now positive toward center of Galaxy)
    u = -u

    if (lsr is not None):   
        u = u + lsr[0]
        v = v + lsr[1]
        w = w + lsr[2]

    return (u,v,w)



import matplotlib
# matplotlib.use('TkAgg')

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
try:
    from matplotlib.backends.backend_tkagg import NavigationToolbar2TkAgg
except ImportError:
    from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk as NavigationToolbar2TkAgg
from matplotlib.figure import Figure
from PyAstronomy.pyaC import pyaErrors as PE

import six.moves.tkinter as tk
import pickle


class Point():

    def __init__(self, event=None):
        self.xdata = 0.0
        self.ydata = 0.0
        self.mplLine = None
        self.lbIdent = None
        if event is not None:
            self.xdata = event.xdata
            self.ydata = event.ydata

    def asStr(self):
        """
        String representation of the point.

        Returns
        -------
        Point : string
            String representation of the point coordinates.
        """
        return "%g, %g" % (self.xdata, self.ydata)


class Picker:
    """
    GUI for interactive point selection.

    Parameters
    ----------
    saveFile : string, optional
        If a filename is given, the list of points will be
        saved to a file with that name using pickle when the
        class instance if destructed. The points are saved as
        a python list of two-float-tuples, holding the x and y
        coordinates of the points. 

    Attributes
    ----------
    f : mpl Figure
        A Figure instance from matplotlib.
    a : mpl axis
        An Axis instance from matplotlib
    pointList : list of points
        A list containing the selected points in the form
        of `Point` instances.
    astyle : string
        Mpl style for plotting used for "active" point.
        Default is "ro".
    istyle : string
        Mpl style for plotting used for "inactive" point.
        Default is "yp".
    """

    def __del__(self):
        """
        Destruct instance
        """
        if not self.saveFile is None:
            # Save current point list to file if filename was given
            pickle.dump(self.points, open(self.saveFile, 'w'))

    def __init__(self, saveFile=None):
        self.windowTitle = "PyA Picker"
        self.f = Figure()
        self.a = self.f.add_subplot(111)

        # Store save filename used to dump data on
        # destruction of class instance
        self.saveFile = saveFile

        # Active and Inactive plot style
        self.astyle = "ro"
        self.istyle = "yp"

        self.pointList = []

        self.root = tk.Tk()

        # A frame containing the mpl plot
        self.plotFrame = tk.Frame()
        self.plotFrame.pack(fill=tk.BOTH, side=tk.LEFT, expand=True)
        self.canvas = FigureCanvasTkAgg(self.f, master=self.plotFrame)

        # A frame containing the box with selected points
        # and control buttons
        self.pointFrame = tk.Frame(self.root)
        self.pointFrame.pack(side=tk.LEFT, fill=tk.BOTH)

        self.listLabel = tk.Label(self.pointFrame, text="Selected points")
        self.listLabel.pack(side=tk.TOP)

        self.lb = tk.Listbox(self.pointFrame, height=15)
        self.lb.pack(side=tk.TOP)
        self.lb.bind("<<ListboxSelect>>", self._lbSelect)

        self.copyButton = tk.Button(
            self.pointFrame, text="Open List", command=self._showList)
        self.copyButton.pack(side=tk.TOP, fill=tk.X)

        self.removeButton = tk.Button(master=self.pointFrame, text="Remove",
                                      command=self._removeButtonClicked)
        self.removeButton.pack(side=tk.TOP, fill=tk.X)

        # a tk.DrawingArea
        self.canvas.get_tk_widget().pack()
        self.cid = self.f.canvas.mpl_connect(
            'button_press_event', self._mouseButtonClicked)

        self.toolbar = NavigationToolbar2TkAgg(self.canvas, self.plotFrame)
        self.toolbar.update()
        self.canvas._tkcanvas.pack(side=tk.TOP, fill=tk.BOTH, expand=True)

        def _quit():
            # stops main loop
            self.root.quit()
            # this is necessary on Windows to prevent
            # Fatal Python Error: PyEval_RestoreThread: NULL tstate
            self.root.destroy()

        self.quitButton = tk.Button(
            master=self.pointFrame, text='Quit', command=_quit)
        self.quitButton.pack(side=tk.BOTTOM, fill=tk.X)

    def _showList(self):
        """
        Creates a new window containing a copy-and-paste aware list of the points.
        """
        def _quitWin():
            win.destroy()
        win = tk.Tk()
        text = tk.Text(master=win)
        text.pack()
        for p in self.pointList:
            text.insert(tk.END, p.asStr() + "\n")
        quit = tk.Button(master=win, text='Quit', command=_quitWin)
        quit.pack()
        win.wm_title("Copy Window")

    def _plotPointAs(self, state, point=None, lbString=None):
        """
        Plot a point in active/inactive color

        Parameters
        ----------
        state : string, {"active", "inactive"}
            Which color to use
        point : Point, optional
            The point information as an instance of the Point class.
        lbString : string
            The point identifier used in the listbox.
        """
        if (point is None) == (lbString is None):
            raise(PE.PyAValError("Either `point` or `lbString` have to be specified."))
        if point is not None:
            pli, lli = self._searchPoint(point.lbIdent)
        else:
            pli, lli = self._searchPoint(lbString)
        # Remove 'old' point from plot
        try:
            self.a.lines.pop(lli)
        except AttributeError:
            self.a.lines[lli].remove()
        if state == "active":
            style = self.astyle
        elif state == "inactive":
            style = self.istyle
        # Add new point in specified color
        self.a.plot([self.pointList[pli].xdata], [
                    self.pointList[pli].ydata], style)
        self.pointList[pli].mplLine = self.a.lines[-1]

    def _searchPoint(self, lbString):
        """
        Search point specified by listbox string. 

        Parameters
        ----------
        lbString : string
            The string used as identifier in the listbox.

        Returns
        -------
        Point list index : int
            The index of the point in the `pointlist`.
        Line index : int
            The index of the corresponding line in the
            lines list attribute of the axis instance. 
        """
        for ip, p in enumerate(self.pointList):
            if p.lbIdent == lbString:
                break
        for i, l in enumerate(self.a.lines):
            if l is p.mplLine:
                break
        return ip, i

    def _lbSelect(self, event):
        """
        React on change of listbox selection.

        This function determines the selected item
        and plots the point in "active" style.
        """
        # Get 'old' active item (changed afterward)
        oa = self.lb.get(tk.ACTIVE)
        # Get active item (string)
        ai = self.lb.get(int(self.lb.curselection()[0]))
        self._plotPointAs("inactive", lbString=oa)
        self._plotPointAs("active", lbString=ai)
        # Update view
        self._updateView()

    def _mouseButtonClicked(self, event):
        """
        Called on click of mouse button.

        If the middle button has been clicked, a
        point is added to the selection.
        """
        # Accept only middle button
        if event.button != 2:
            return
        # Check whether click occurred on plot axes.
        if event.inaxes is not self.a:
            return

        p = Point(event)
        self.a.plot([p.xdata], [p.ydata], self.istyle)
        p.mplLine = self.a.lines[-1]
        self.f.canvas.draw()

        # Insert the point's content into the list box
        self.lb.insert(tk.END, p.asStr())
        # Save the "list box identifier"
        p.lbIdent = p.asStr()
        # Save the point
        self.pointList.append(p)

    def _removeButtonClicked(self):
        """
        Remove a point.
        """
        # Check whether anything was selected
        sel = self.lb.curselection()
        if len(sel) == 0:
            return
        # Find indices for the point
        el = self.lb.get(int(sel[0]))
        pli, lli = self._searchPoint(el)
        # Remove it
        try:
            self.a.lines.pop(lli)
        except AttributeError:
            self.a.lines[lli].remove()
        self.pointList.pop(pli)
        self.lb.delete(sel[0])
        self._updateView()

    def _updateView(self):
        """
        Redraw MPL canvas
        """
        self.f.canvas.draw()

    def pick(self):
        """
        Select points.

        .. note:: Before selection points, a plot should
                  have been created from which points may
                  be selected.

        Returns
        -------
        Points : List of tuples
            A list of two-element tuples, holding the x and
            y coordinates of the selected points.
        """
        self.root.wm_title(self.windowTitle)
        try:
            self.canvas.show()
        except AttributeError:
            # Seems to be dispensable with newer (> about 3) matplotlib versions
            pass
        tk.mainloop()
        # Prepare return value
        self.points = []
        for p in self.pointList:
            self.points.append((p.xdata, p.ydata))
        return self.points[:]



from __future__ import print_function, division
import numpy
from PyAstronomy.pyaC import pyaErrors as PE


def aitoff(l, b):
    """
    Carry out Aitoff projection. 

    Parameters
    ----------
    l, b : float or array
        The longitude and latitude [deg]

    Returns
    -------
    x, y : float or array
        Aitoff-projected coordinates (`x`, `y`)

    Notes
    -----

    .. note:: This function was ported from the IDL Astronomy User's Library.

    :IDL - Documentation:

    pro aitoff,l,b,x,y
    +
     NAME:
           AITOFF
     PURPOSE:
           Convert longitude, latitude to X,Y using an AITOFF projection.
     EXPLANATION:
           This procedure can be used to create an all-sky map in Galactic 
           coordinates with an equal-area Aitoff projection.  Output map 
           coordinates are zero longitude centered.

     CALLING SEQUENCE:
           AITOFF, L, B, X, Y 

     INPUTS:
           L - longitude - scalar or vector, in degrees
           B - latitude - same number of elements as L, in degrees

     OUTPUTS:
           X - X coordinate, same number of elements as L.   X is normalized to
                   be between -180 and 180
           Y - Y coordinate, same number of elements as L.  Y is normalized to
                   be between -90 and 90.

     NOTES:
           See AIPS memo No. 46, page 4, for details of the algorithm.  This
           version of AITOFF assumes the projection is centered at b=0 degrees.

     REVISION HISTORY:
           Written  W.B. Landsman  STX          December 1989
           Modified for Unix:
                   J. Bloch        LANL SST-9      5/16/91 1.1
           Converted to IDL V5.0   W. Landsman   September 1997
    """
    wasFloat = False
    if isinstance(l, float):
        l = numpy.array([l])
        b = numpy.array([b])
        wasFloat = True

    sa = l.copy()
    x180 = numpy.where(sa > 180.0)[0]
    if len(x180) > 0:
        sa[x180] -= 360.
    alpha2 = sa/(2*(180.0/numpy.pi))
    delta = b/(180.0/numpy.pi)
    r2 = numpy.sqrt(2.)
    f = 2*r2/numpy.pi
    cdec = numpy.cos(delta)
    denom = numpy.sqrt(1. + cdec*numpy.cos(alpha2))
    x = cdec*numpy.sin(alpha2)*2.*r2/denom
    y = numpy.sin(delta)*r2/denom
    x = x*(180.0/numpy.pi)/f
    y = y*(180.0/numpy.pi)/f

    if wasFloat:
        return float(x[0]), float(y[0])
    else:
        return x, y


def inverseAitoff(x, y):
    """
    Carry out an inverse Aitoff projection.

    This function reverts to aitoff projection made by the
    function *aitoff*. The result is either two floats or
    arrays (depending on whether float or array was used
    as input) representing longitude and latitude. Both are
    given in degrees with -180 < longitude < +180 and
    -90 < latitude < 90.

    Parameters
    ----------
    x : float or array
        A value between -180. and +180.
        (see convention in *aitoff* function).
    y : float or array
        A value between -90. and +90
        (see convention in *aitoff* function).

    Returns
    -------
    Deprojected coordinates : float or array
        If arrays are used for input, the function returns an
        array for the longitude, for the latitude, and
        an index array containing those array indices for which
        the reprojection could be carried out.
    """
    wasFloat = False
    if isinstance(x, float):
        x = numpy.array([x])
        y = numpy.array([y])
        wasFloat = True

    # First, rescale x and y
    x = x/180.0 * 2.0 * numpy.sqrt(2.0)
    y = y/90.0 * numpy.sqrt(2.0)

    zsqr = 1.0 - (x/4.0)**2 - (y/2.0)**2
    # Check whether x,y coordinates are within the ellipse of invertible values.
    indi = numpy.where((x**2/8. + y**2/2. - 1.0) <= 0.0)[0]
    if len(indi) == 0:
        raise (PE.PyAValError("Deprojection is not possible.",
               where="inverseAitoff", why="No values inside valid space."))

    z = numpy.sqrt(zsqr[indi])
    l = 2.0 * numpy.arctan(z*x[indi]/(2.0 * (2.0*z**2 - 1.0)))
    b = numpy.arcsin(z * y[indi])

    l = l*180.0 / numpy.pi
    b = b*180.0 / numpy.pi

    if wasFloat:
        return float(l[0]), float(b[0])
    else:
        return l, b, indi



import numpy as np
import scipy.interpolate as interpolate

def unred(wave, flux, ebv, R_V=3.1, LMC2=False, AVGLMC=False):
    """
     Deredden a flux vector using the Fitzpatrick (1999) parameterization
     
     Parameters
     ----------
     wave :   array
              Wavelength in Angstrom
     flux :   array
              Calibrated flux vector, same number of elements as wave.
     ebv  :   float, optional
              Color excess E(B-V). If a negative ebv is supplied,
              then fluxes will be reddened rather than dereddened.
              The default is 3.1.
     AVGLMC : boolean
              If True, then the default fit parameters c1,c2,c3,c4,gamma,x0 
              are set to the average values determined for reddening in the 
              general Large Magellanic Cloud (LMC) field by
              Misselt et al. (1999, ApJ, 515, 128). The default is
              False.
     LMC2 :   boolean
              If True, the fit parameters are set to the values determined
              for the LMC2 field (including 30 Dor) by Misselt et al.
              Note that neither `AVGLMC` nor `LMC2` will alter the default value 
              of R_V, which is poorly known for the LMC.
       
     Returns
     -------             
     new_flux : array 
                Dereddened flux vector, same units and number of elements
                as input flux.
     
     Notes
     -----
    
     .. note:: This function was ported from the IDL Astronomy User's Library.
    
     :IDL - Documentation:
     
      PURPOSE:
       Deredden a flux vector using the Fitzpatrick (1999) parameterization
      EXPLANATION:
       The R-dependent Galactic extinction curve is that of Fitzpatrick & Massa 
       (Fitzpatrick, 1999, PASP, 111, 63; astro-ph/9809387 ).    
       Parameterization is valid from the IR to the far-UV (3.5 microns to 0.1 
       microns).    UV extinction curve is extrapolated down to 912 Angstroms.
    
      CALLING SEQUENCE:
        FM_UNRED, wave, flux, ebv, [ funred, R_V = , /LMC2, /AVGLMC, ExtCurve= 
                          gamma =, x0=, c1=, c2=, c3=, c4= ]
      INPUT:
         WAVE - wavelength vector (Angstroms)
         FLUX - calibrated flux vector, same number of elements as WAVE
                  If only 3 parameters are supplied, then this vector will
                  updated on output to contain the dereddened flux.
         EBV  - color excess E(B-V), scalar.  If a negative EBV is supplied,
                  then fluxes will be reddened rather than dereddened.
    
      OUTPUT:
         FUNRED - unreddened flux vector, same units and number of elements
                  as FLUX
    
      OPTIONAL INPUT KEYWORDS
          R_V - scalar specifying the ratio of total to selective extinction
                   R(V) = A(V) / E(B - V).    If not specified, then R = 3.1
                   Extreme values of R(V) range from 2.3 to 5.3
    
       /AVGLMC - if set, then the default fit parameters c1,c2,c3,c4,gamma,x0 
                 are set to the average values determined for reddening in the 
                 general Large Magellanic Cloud (LMC) field by Misselt et al. 
                 (1999, ApJ, 515, 128)
        /LMC2 - if set, then the fit parameters are set to the values determined
                 for the LMC2 field (including 30 Dor) by Misselt et al.
                 Note that neither /AVGLMC or /LMC2 will alter the default value 
                 of R_V which is poorly known for the LMC. 
                
         The following five input keyword parameters allow the user to customize
         the adopted extinction curve.    For example, see Clayton et al. (2003,
         ApJ, 588, 871) for examples of these parameters in different interstellar
         environments.
    
         x0 - Centroid of 2200 A bump in microns (default = 4.596)
         gamma - Width of 2200 A bump in microns (default  =0.99)
         c3 - Strength of the 2200 A bump (default = 3.23)
         c4 - FUV curvature (default = 0.41)
         c2 - Slope of the linear UV extinction component 
              (default = -0.824 + 4.717/R)
         c1 - Intercept of the linear UV extinction component 
              (default = 2.030 - 3.007*c2
    """

    x = 10000./ wave # Convert to inverse microns 
    curve = x*0.
    
    # Set some standard values:
    x0 = 4.596
    gamma =  0.99
    c3 =  3.23      
    c4 =  0.41    
    c2 = -0.824 + 4.717/R_V
    c1 =  2.030 - 3.007*c2
    
    if LMC2:
        x0    =  4.626
        gamma =  1.05   
        c4   =  0.42   
        c3    =  1.92      
        c2    = 1.31
        c1    =  -2.16
    elif AVGLMC:   
        x0 = 4.596  
        gamma = 0.91
        c4   =  0.64  
        c3    =  2.73      
        c2    = 1.11
        c1    =  -1.28
    
    # Compute UV portion of A(lambda)/E(B-V) curve using FM fitting function and 
    # R-dependent coefficients
    xcutuv = np.array([10000.0/2700.0])
    xspluv = 10000.0/np.array([2700.0,2600.0])
    
    iuv = np.where(x >= xcutuv)[0]
    N_UV = len(iuv)
    iopir = np.where(x < xcutuv)[0]
    Nopir = len(iopir)
    if (N_UV > 0): xuv = np.concatenate((xspluv,x[iuv]))
    else:  xuv = xspluv
    
    yuv = c1  + c2*xuv
    yuv = yuv + c3*xuv**2/((xuv**2-x0**2)**2 +(xuv*gamma)**2)
    yuv = yuv + c4*(0.5392*(np.maximum(xuv,5.9)-5.9)**2+0.05644*(np.maximum(xuv,5.9)-5.9)**3)
    yuv = yuv + R_V
    yspluv  = yuv[0:2]  # save spline points
     
    if (N_UV > 0): curve[iuv] = yuv[2::] # remove spline points
    
    # Compute optical portion of A(lambda)/E(B-V) curve
    # using cubic spline anchored in UV, optical, and IR
    xsplopir = np.concatenate(([0],10000.0/np.array([26500.0,12200.0,6000.0,5470.0,4670.0,4110.0])))
    ysplir   = np.array([0.0,0.26469,0.82925])*R_V/3.1 
    ysplop   = np.array((np.polyval([-4.22809e-01, 1.00270, 2.13572e-04][::-1],R_V ), 
            np.polyval([-5.13540e-02, 1.00216, -7.35778e-05][::-1],R_V ), 
            np.polyval([ 7.00127e-01, 1.00184, -3.32598e-05][::-1],R_V ), 
            np.polyval([ 1.19456, 1.01707, -5.46959e-03, 7.97809e-04, -4.45636e-05][::-1],R_V ) ))
    ysplopir = np.concatenate((ysplir,ysplop))
    
    if (Nopir > 0): 
      tck = interpolate.splrep(np.concatenate((xsplopir,xspluv)),np.concatenate((ysplopir,yspluv)),s=0)
      curve[iopir] = interpolate.splev(x[iopir], tck)
    
    #Now apply extinction correction to input flux vector
    curve *= ebv
    
    return flux * 10.**(0.4*curve)



from __future__ import print_function, division
from PyAstronomy.pyaC import pyaPermanent as pp
from PyAstronomy.pyaC import pyaErrors as PE
import os
import gzip
from PyAstronomy.pyasl import _ic
import ssl


class SWEETCat(pp.PyAUpdateCycle):
    """
    Access the SWEET-Cat catalog.

    The SWEET-Cat catalog provides parameters for planet host stars.

    The following data are provided

    ===========   ===================================  ======
    Column        Description                          Unit
    -----------   -----------------------------------  ------
    star          Name of the star
    hd            The HD number (if available)
    ra            The right ascension                   hms
    dec           The declination                       dms
    vmag          V magnitude                           mag
    ervmag        Error on V magnitude                  mag
    par           Parallax                              mas
    erpar         Error on parallax                     mas
    parsource     If par is from Simbad or calculated
    teff          Effective temperature                 K
    erteff        Error on effective temperature        K
    logg          Surface gravity                       cgs
    erlogg        Error on surface gravity              cgs
    logglc        Surface gravity from LC               cgs
    erlogglc      Error on surface gravity from LC      cgs
    vt            Micro turbulence                      km/s
    ervt          Error on micro turbulence             km/s
    metal         Metallicity ([Fe/H])
    ermetal       Error on metallicity
    mass          Mass calculated from Torres et al.    Solar
    ermass        Error on calculated mass              Solar
    author        Author of source
    link          Link to paper in ADS
    source        1 means CAUP's method. 0 otherwise
    update        When the parameters were updated
    comment1      Special comment
    comment2      Blank
    ===========   ===================================  ======


    Detailed information can be found here:
    https://www.astro.up.pt/resources/sweet-cat/
    and in the associated publications.

    Attributes
    ----------
    data : pandas data frame
        The catalog data

    """

    def _downloadData(self):
        """
        Download SWEETCAT and write it to file
        """
        urls = ['https://www.astro.up.pt/resources/sweet-cat/download.php', \
                "https://raw.githubusercontent.com/iastro-pt/SWEET-Cat/master/WEBSITE_online_EU.rdb"]
        dfn = self._fs.composeFilename(self.dataFileName)
        success = False
        ex = []
        for url in urls:
            try:
                # The next two lines by-pass certificate verification
                context = ssl._create_unverified_context()
                self._fs.downloadToFile(url, dfn, clobber=True, verbose=False, openMethod=gzip.open,
                                        context=context)
                success = True
            except AttributeError:
                # Python version does not support ssl._create_unverified_context()
                self._fs.downloadToFile(
                    url, dfn, clobber=True, verbose=False, openMethod=gzip.open)
                success = True
            except Exception as e:
                # "Handle" unexpected error
                ex.append((url,e))
            if success:
                break
        
        if not success:
            raise(PE.PyANetworkError(
                    "Could not download SWEET-Cat data. Received the following errors:\n\n" + \
                    ''.join(["URL: "+x[0]+"\n"+str(x[1])+"\n\n" for x in ex]) ))

        self.source_url = url

    def _read_sweetcat(self):
        """
        Read SWEETCat into a pandas DataFrame
        """
        if not _ic.check["pandas"]:
            raise(PE.PyARequiredImport("Could not import pandas module.",
                                       solution="Please install pandas (http://pandas.pydata.org). Please also check the dependencies."))
        else:
            import pandas as pd
        ffn = self._fs.requestFile(self.dataFileName, 'r', gzip.open)
        self.data = pd.read_csv(
            ffn, sep='\t', names=self.names, na_values=['~'])

    def __init__(self, skipUpdate=False):
        self.dataFileName = os.path.join(
            "pyasl", "resBased", "sweetcat.csv.gz")
        configFilename = os.path.join("pyasl", "resBased", "sweetcat.cfg")
        pp.PyAUpdateCycle.__init__(self, configFilename, "sweetcatupdate")

        self.names = ['star', 'hd', 'ra', 'dec', 'vmag', 'ervmag', 'par', 'erpar',
                      'parsource', 'teff', 'erteff', 'logg', 'erlogg', 'logglc',
                      'erlogglc', 'vt', 'ervt', 'metal', 'ermetal', 'mass', 'ermass',
                      'author', 'source', 'update', 'comment']

        # Check whether data file exists
        self._fs = pp.PyAFS()
        if (self.needsUpdate() or (not self._fs.fileExists(self.dataFileName))) and (not skipUpdate):
            # Data needs update
            print("Downloading exoplanet data from SWEET-Cat archive")
            self._update(self._downloadData)
            print("Saved data to file: ",
                  self.dataFileName, " in data directory,")
            print("  which has been configured as: ", self._fs.dpath)
            print("By default, the data will be downloaded anew every 7 days.")
            print("You can use the `changeDownloadCycle` to change this behavior.")
        self._read_sweetcat()

    def downloadData(self):
        """
        Trigger download of data.
        """
        self._update(self._downloadData)
        self._read_sweetcat()



# -*- coding: utf-8 -*-
from __future__ import print_function, division
import numpy as np
from .onedfit import OneDFit
from PyAstronomy.pyaC import pyaErrors as PE


class Sech1d(OneDFit):
    r"""
    Implements a one dimensional hyperbolic secant

    The functional form is:

    .. math:: \frac{2 A}{e^{(x-mu)/w} + e^{-(x-mu)/w} } + x \times lin + off

    Here, `lin` and `off` denote the linear and the offset term.

    .. note:: The area under the curve is given by :math:`\pi \times A`

    *Fit parameters*:
     - `A` - Amplitude (maximum/minimum of the curve, not area)
     - `mu` - Center of the hyperbolic secant
     - `w` - Width parameter
     - `off` - Offset
     - `lin` - Linear term
    """

    def __init__(self):
        OneDFit.__init__(self, ["A", "mu", "w", "off", "lin"])
        self.setRootName("Sech")

    def evaluate(self, x):
        r"""
        Evaluates the model for current parameter values.

        Parameters
        ----------
        x : array
            Specifies the points at which to evaluate the model.
        """
        if self["w"] == 0.0:
            raise (PE.PyAValError("Width of Sech must be larger than zero.",
                                  solution="Change width ('w')."))
        z = (x-self["mu"]) / self["w"]
        y = (2. * self["A"]) / (np.exp(z) + np.exp(-z)) + \
            + self["off"] + (self["lin"] * x)
        return y



from PyAstronomy import pyaC
from PyAstronomy.pyasl import _moduleImportStatus

_modules = ["baraffe98tracks", "nasaExoplanetArchive", "exoplanetEU", \
           "exoplanetsOrg", "kuruczModels", "fip", "sweet_cat", "spectralLib"]

for m in _modules:
  _moduleImportStatus[(__name__, m)] = pyaC.pyaimportallfrom(m, "PyAstronomy.pyasl.resBased", globals())

# from baraffe98tracks import *
# from nasaExoplanetArchive import *
# from exoplanetEU import *
# from exoplanetsOrg import *
# from kuruczModels import *
# from fip import *
# from sweet_cat import *


from __future__ import print_function, division
import numpy as np
from .onedfit import OneDFit


class Circle2d(OneDFit):
    r"""
    Implements a simple, two-dimensional circular orbit.

    The functional form is:

    .. math:: (x,y) = (r \cos(\omega (t-t_0)), r \sin(\omega (t-t_0)))

    where :math:`\omega` is :math:`2\pi/Period`.

    *Fit parameters*:
     - `r` - Radius of the circle
     - `per` - Period of orbit
     - `t0` - Starting time of orbit
    """

    def __init__(self):
        OneDFit.__init__(self, ["r", "per", "t0"])
        self.setRootName("Circle2d")

    def evaluate(self, t):
        r"""
        Evaluates the model for current parameter values.

        Parameters
        ----------
        x : array
            Specifies the points at which to evaluate the model.
        """
        result = np.zeros((len(t), 2))
        w = 2.0*np.pi / self["per"]
        result[::, 0] = abs(self["r"]) * np.cos(w*(t-self["t0"]))
        result[::, 1] = abs(self["r"]) * np.sin(w*(t-self["t0"]))
        return result



from __future__ import print_function
import copy
import six
import traceback

class PyaErrTemplate(Exception):
  
  def __init__(self, what, errType, where=None, why=None, solution=None, addInfo=None, tbfe=None):
    """
      The PyA error class template.
      
      Parameters
      ----------
      what : string
          What has happened?
      errType : string
          Description of the type of error.
      where : string
          Where did it happen?
      why : string
          Why did it happen?
      solution : string or list of strings
          How can the problem be solved?
      addInfo : string
          Whatever additional information may be available.
      tbfe : Exception
          Saves trace back from a previously raised exception.
      
      Of the parameters, only the first (what) and second (errType) are mandatory; the latter
      should be provided by every more specialized exception class.
      
      .. note:: The template should never be raised directly, but specialized derived classes
                should be used.
    """
    self.errorType = errType
    self.what = what
    self.where = where
    self.why = why
    # Check whether solution is iterable
    self.solution = solution
    if solution is not None:
      if not isinstance(solution, six.string_types):
        self.solution = copy.deepcopy(solution)
      else:
        self.solution = [solution]    
    self.addInfo = addInfo
    # Check trace back
    if not tbfe is None:
      self.addTB(tbfe)
    else:
      self.tbfe = None
  
  def addTB(self, e):
    """
      Add trace back from another exception.
      
      Parameters
      ----------
      e : Exception
          The exception from which to add.
    """
    self.tbfe = ["  - " + l for l in traceback.format_exc().splitlines(True)]
  
  def __str__(self, head=True):
    """
      Return error message as string.
      
      Parameters:
        - `head` - boolean, If True, a header will be printed says "PyA error". Setting False is, e.g.,
                   used in printing warnings.
    """
    e = ""
    if head:
      e = "\n"
      e += "---------------------\n"
      e += "A PyA error occurred:\n"
      e += "---------------------\n"
    if not self.tbfe is None:
      # Print information from trace back (add indentation)
      e += "PyA trace back:\n"
      for l in self.tbfe:
        e += l
    if head:
      e += "Type of error: " + self.errorType + "\n"
    e += "What happened?\n"
    e += "    " + self.what + "\n"
    if not self.where is None:
      e += "Where did it happen?\n"
      e += "    " + self.where + "\n"
    if not self.why is None:
      e += "Why did it happen?\n"
      e += "    " + self.why + "\n"
    if not self.solution is None:
      e += "What are possible solutions?\n"
      for s in self.solution:
        e += "  - " + s + "\n"
    if not self.addInfo is None:
      e += "Additional information:\n"
      e += "    " + self.addInfo + "\n"
    return e



import unittest
class TestSanityOfNotebook(unittest.TestCase):
    def setUp(self):
        pass
    def tearDown(self):
        pass
        
    def testsanity_example(self):
        #!/usr/bin/env python
        # coding: utf-8
        
        # Restrictions via penalties
        # ===============================
        
        # In[1]:
        
        
        from numpy import arange, sqrt, exp, pi, random, ones_like
        import numpy as np
        import matplotlib.pylab as plt
        from PyAstronomy import funcFit2 as fuf2
        import scipy.optimize as sco
        import matplotlib
        matplotlib.use('Agg')
        
        random.seed(1234)
        
        # Creating a Gaussian with some noise
        # Choose some parameters...
        gPar = {"A":1.0, "sig":10.0, "mu":10.0, "off":1.0, "lin":0.0}
        # Calculate profile
        x = arange(100) - 50.0
        y = gPar["off"] + gPar["A"] / sqrt(2*pi*gPar["sig"]**2) \
            * exp(-(x-gPar["mu"])**2/(2*gPar["sig"]**2))
        # Add some noise...
        y += random.normal(0.0, 0.002, x.size)
        # ...and save the error bars
        yerr = ones_like(x)*0.002
        
        # Create a model object
        gf = fuf2.GaussFit()
        
        # Set guess values for the parameters
        gf.assignValues({"A":3, "sig":3.77, "off":0.96, "mu":9.5})
        
        # 'Thaw' those (the order is irrelevant)
        gf.thaw(["mu", "sig", "off", "A"])
        
        # Restrict parameter range of sigma to within 4.1 to 7.1
        # where the upper limit is too small for input values
        gf.setRestriction({"sig":[4.1,7.1]})
        # An upper limit of 10 with no lower limit on the area of the Gaussian
        gf.setRestriction({"A":[None,10]})
        
        # Check restrictions 
        print("Parameter restrictions: ", gf.getRestrictions())
        print()
        print("Penalty factor: ", gf.getPenaltyFactor())
        
        # Inspect the effect of the restriction on the objective function value
        sigs = np.linspace(3,8,100)
        oval = np.zeros_like(sigs)
        
        for i, sig in enumerate(sigs):
            gf["sig"] = sig
            oval[i] = gf.chisqr(gf.freeParamVals(),x,y,yerr)
            
        plt.title("Value of objective function (with 'penalty ramps')")
        plt.xlabel("Sigma")
        plt.plot(sigs, oval, 'b.-')
        plt.show()
        
        # Use a minimization algorithm, which itself ignores
        # restrictions, and observe effect of penalties
        fr = fuf2.fitfmin(gf, gf.chisqr, x, y, yerr=yerr)
        print()
        print("Fit result: ", fr)
        print()
        
        gf.parameterSummary()
        plt.title("Bad fit caused by restrictions")
        plt.plot(x, y, 'bp')
        plt.plot(x, gf.evaluate(x), 'r--')
        plt.show()
        



from __future__ import division
import numpy


class SaniBase:
    def mrd(self, x, y):
        """
        Returns maximum relative deviation = max(abs(x-y)/x)

        Parameters:
         - `x`, `y` - float or array
        """
        return numpy.max(numpy.abs(x - y) / x)



"""
    Python implementation of the routines below

    subroutine occultquad(z0,u1,u2,p,muo1,mu0,nz)
C  This routine computes the lightcurve for occultation
C  of a quadratically limb-darkened source without microlensing.
C  Please cite Mandel & Agol (2002) if you make use of this routine
C  in your research.  Please report errors or bugs to agol@tapir.caltech.edu
      implicit none
      integer i,nz
      double precision z0(nz),u1,u2,p,muo1(nz),mu0(nz),
     &       mu(nz),lambdad(nz),etad(nz),lambdae(nz),lam,
     &       pi,x1,x2,x3,z,omega,kap0,kap1,q,Kk,Ek,Pk,n,ellec,ellk,rj
      if(abs(p-0.5d0).lt.1.d-3) p=0.5d0
C
C Input:
C
C rs   radius of the source (set to unity)
C z0   impact parameter in units of rs
C p    occulting star size in units of rs
C u1   linear    limb-darkening coefficient (gamma_1 in paper)
C u2   quadratic limb-darkening coefficient (gamma_2 in paper)
C
C Output:
C
C muo1 fraction of flux at each z0 for a limb-darkened source
C mu0  fraction of flux at each z0 for a uniform source
C
C Limb darkening has the form:
C  I(r)=[1-u1*(1-sqrt(1-(r/rs)^2))-u2*(1-sqrt(1-(r/rs)^2))^2]/(1-u1/3-u2/6)/pi
C 
C To use this routine
C
C Now, compute pure occultation curve:
      omega=1.d0-u1/3.d0-u2/6.d0
      pi=acos(-1.d0)
C Loop over each impact parameter:
      do i=1,nz
C substitute z=z0(i) to shorten expressions
        z=z0(i)
        x1=(p-z)**2
        x2=(p+z)**2
        x3=p**2-z**2
C the source is unocculted:
C Table 3, I.
        if(z.ge.1.d0+p) then
          lambdad(i)=0.d0
          etad(i)=0.d0
          lambdae(i)=0.d0
          goto 10
        endif
C the  source is completely occulted:
C Table 3, II.
        if(p.ge.1.d0.and.z.le.p-1.d0) then
          lambdad(i)=1.d0
          etad(i)=1.d0
          lambdae(i)=1.d0
          goto 10
        endif
C the source is partly occulted and the occulting object crosses the limb:
C Equation (26):
        if(z.ge.abs(1.d0-p).and.z.le.1.d0+p) then
          kap1=acos(min((1.d0-p*p+z*z)/2.d0/z,1.d0))
          kap0=acos(min((p*p+z*z-1.d0)/2.d0/p/z,1.d0))
          lambdae(i)=p*p*kap0+kap1
          lambdae(i)=(lambdae(i)-0.5d0*sqrt(max(4.d0*z*z-
     &               (1.d0+z*z-p*p)**2,0.d0)))/pi
        endif
C the occulting object transits the source star (but doesn't
C completely cover it):
        if(z.le.1.d0-p) lambdae(i)=p*p
C the edge of the occulting star lies at the origin- special 
C expressions in this case:
        if(abs(z-p).lt.1.d-4*(z+p)) then
C Table 3, Case V.:
          if(z.ge.0.5d0) then
            lam=0.5d0*pi
            q=0.5d0/p
            Kk=ellk(q)
            Ek=ellec(q)
C Equation 34: lambda_3
            lambdad(i)=1.d0/3.d0+16.d0*p/9.d0/pi*(2.d0*p*p-1.d0)*Ek-
     &                 (32.d0*p**4-20.d0*p*p+3.d0)/9.d0/pi/p*Kk
C Equation 34: eta_1
            etad(i)=1.d0/2.d0/pi*(kap1+p*p*(p*p+2.d0*z*z)*kap0-
     &              (1.d0+5.d0*p*p+z*z)/4.d0*sqrt((1.d0-x1)*(x2-1.d0)))
            if(p.eq.0.5d0) then
C Case VIII: p=1/2, z=1/2
              lambdad(i)=1.d0/3.d0-4.d0/pi/9.d0
              etad(i)=3.d0/32.d0
            endif
            goto 10
          else
C Table 3, Case VI.:
            lam=0.5d0*pi
            q=2.d0*p
            Kk=ellk(q)
            Ek=ellec(q)
C Equation 34: lambda_4
            lambdad(i)=1.d0/3.d0+2.d0/9.d0/pi*(4.d0*(2.d0*p*p-1.d0)*Ek+
     &                 (1.d0-4.d0*p*p)*Kk)
C Equation 34: eta_2
            etad(i)=p*p/2.d0*(p*p+2.d0*z*z)
            goto 10
          endif
        endif
C the occulting star partly occults the source and crosses the limb:
C Table 3, Case III:
        if((z.gt.0.5d0+abs(p-0.5d0).and.z.lt.1.d0+p).or.(p.gt.0.5d0.
     &      and.z.gt.abs(1.d0-p)*1.0001d0.and.z.lt.p)) then
          lam=0.5d0*pi
          q=sqrt((1.d0-(p-z)**2)/4.d0/z/p)
          Kk=ellk(q)
          Ek=ellec(q)
          n=1.d0/x1-1.d0
          Pk=Kk-n/3.d0*rj(0.d0,1.d0-q*q,1.d0,1.d0+n)
C Equation 34, lambda_1:
          lambdad(i)=1.d0/9.d0/pi/sqrt(p*z)*(((1.d0-x2)*(2.d0*x2+
     &        x1-3.d0)-3.d0*x3*(x2-2.d0))*Kk+4.d0*p*z*(z*z+
     &        7.d0*p*p-4.d0)*Ek-3.d0*x3/x1*Pk)
          if(z.lt.p) lambdad(i)=lambdad(i)+2.d0/3.d0
C Equation 34, eta_1:
          etad(i)=1.d0/2.d0/pi*(kap1+p*p*(p*p+2.d0*z*z)*kap0-
     &          (1.d0+5.d0*p*p+z*z)/4.d0*sqrt((1.d0-x1)*(x2-1.d0)))
          goto 10
        endif
C the occulting star transits the source:
C Table 3, Case IV.:
        if(p.le.1.d0.and.z.le.(1.d0-p)*1.0001d0) then
          lam=0.5d0*pi
          q=sqrt((x2-x1)/(1.d0-x1))
          Kk=ellk(q)
          Ek=ellec(q)
          n=x2/x1-1.d0
          Pk=Kk-n/3.d0*rj(0.d0,1.d0-q*q,1.d0,1.d0+n)
C Equation 34, lambda_2:
          lambdad(i)=2.d0/9.d0/pi/sqrt(1.d0-x1)*((1.d0-5.d0*z*z+p*p+
     &         x3*x3)*Kk+(1.d0-x1)*(z*z+7.d0*p*p-4.d0)*Ek-3.d0*x3/x1*Pk)
          if(z.lt.p) lambdad(i)=lambdad(i)+2.d0/3.d0
          if(abs(p+z-1.d0).le.1.d-4) then
            lambdad(i)=2/3.d0/pi*acos(1.d0-2.d0*p)-4.d0/9.d0/pi*
     &            sqrt(p*(1.d0-p))*(3.d0+2.d0*p-8.d0*p*p)
          endif
C Equation 34, eta_2:
          etad(i)=p*p/2.d0*(p*p+2.d0*z*z)
        endif
 10     continue
C Now, using equation (33):
        muo1(i)=1.d0-((1.d0-u1-2.d0*u2)*lambdae(i)+(u1+2.d0*u2)*
     &      lambdad(i)+u2*etad(i))/omega
C Equation 25:
        mu0(i)=1.d0-lambdae(i)
      enddo
      return
      end

      FUNCTION rc(x,y)
      REAL*8 rc,x,y,ERRTOL,TINY,SQRTNY,BIG,TNBG,COMP1,COMP2,THIRD,C1,C2,
     *C3,C4
      PARAMETER (ERRTOL=.04d0,TINY=1.69d-38,SQRTNY=1.3d-19,BIG=3.d37,
     *TNBG=TINY*BIG,COMP1=2.236d0/SQRTNY,COMP2=TNBG*TNBG/25.d0,
     *THIRD=1.d0/3.d0,C1=.3d0,C2=1.d0/7.d0,C3=.375d0,C4=9.d0/22.d0)
      REAL*8 alamb,ave,s,w,xt,yt
      if(x.lt.0..or.y.eq.0..or.(x+abs(y)).lt.TINY.or.(x+
     *abs(y)).gt.BIG.or.(y.lt.-COMP1.and.x.gt.0..and.x.lt.COMP2))pause 
     *'invalid arguments in rc'
      if(y.gt.0.d0)then
        xt=x
        yt=y
        w=1.
      else
        xt=x-y
        yt=-y
        w=sqrt(x)/sqrt(xt)
      endif
1     continue
        alamb=2.d0*sqrt(xt)*sqrt(yt)+yt
        xt=.25d0*(xt+alamb)
        yt=.25d0*(yt+alamb)
        ave=THIRD*(xt+yt+yt)
        s=(yt-ave)/ave
      if(abs(s).gt.ERRTOL)goto 1
      rc=w*(1.d0+s*s*(C1+s*(C2+s*(C3+s*C4))))/sqrt(ave)
      return
      END
C  (C) Copr. 1986-92 Numerical Recipes Software

      FUNCTION rj(x,y,z,p)
      REAL*8 rj,p,x,y,z,ERRTOL,TINY,BIG,C1,C2,C3,C4,C5,C6,C7,C8
      PARAMETER (ERRTOL=.05d0,TINY=2.5d-13,BIG=9.d11,C1=3.d0/14.d0,
     *C2=1.d0/3.d0,C3=3.d0/22.d0,C4=3.d0/26.d0,C5=.75d0*C3,
     *C6=1.5d0*C4,C7=.5d0*C2,C8=C3+C3)
CU    USES rc,rf
      REAL*8 a,alamb,alpha,ave,b,beta,delp,delx,dely,delz,ea,eb,ec,ed,ee,
     *fac,pt,rcx,rho,sqrtx,sqrty,sqrtz,sum,tau,xt,yt,zt,rc,rf
      if(min(x,y,z).lt.0..or.min(x+y,x+z,y+z,abs(p)).lt.TINY.or.max(x,y,
     *z,abs(p)).gt.BIG)pause 'invalid arguments in rj'
      sum=0.d0
      fac=1.d0
      if(p.gt.0.d0)then
        xt=x
        yt=y
        zt=z
        pt=p
      else
        xt=min(x,y,z)
        zt=max(x,y,z)
        yt=x+y+z-xt-zt
        a=1.d0/(yt-p)
        b=a*(zt-yt)*(yt-xt)
        pt=yt+b
        rho=xt*zt/yt
        tau=p*pt/yt
        rcx=rc(rho,tau)
      endif
1     continue
        sqrtx=sqrt(xt)
        sqrty=sqrt(yt)
        sqrtz=sqrt(zt)
        alamb=sqrtx*(sqrty+sqrtz)+sqrty*sqrtz
        alpha=(pt*(sqrtx+sqrty+sqrtz)+sqrtx*sqrty*sqrtz)**2
        beta=pt*(pt+alamb)**2
        sum=sum+fac*rc(alpha,beta)
        fac=.25d0*fac
        xt=.25d0*(xt+alamb)
        yt=.25d0*(yt+alamb)
        zt=.25d0*(zt+alamb)
        pt=.25d0*(pt+alamb)
        ave=.2d0*(xt+yt+zt+pt+pt)
        delx=(ave-xt)/ave
        dely=(ave-yt)/ave
        delz=(ave-zt)/ave
        delp=(ave-pt)/ave
      if(max(abs(delx),abs(dely),abs(delz),abs(delp)).gt.ERRTOL)goto 1
      ea=delx*(dely+delz)+dely*delz
      eb=delx*dely*delz
      ec=delp**2
      ed=ea-3.d0*ec
      ee=eb+2.d0*delp*(ea-ec)
      rj=3.d0*sum+fac*(1.d0+ed*(-C1+C5*ed-C6*ee)+eb*(C7+delp*
     *(-C8+delp*C4))+delp*ea*(C2-delp*C3)-C2*delp*ec)/(ave*sqrt(ave))
      if (p.le.0.d0) rj=a*(b*rj+3.d0*(rcx-rf(xt,yt,zt)))
      return
      END
C  (C) Copr. 1986-92 Numerical Recipes Software

      function ellec(k)
      implicit none
      double precision k,m1,a1,a2,a3,a4,b1,b2,b3,b4,ee1,ee2,ellec
C Computes polynomial approximation for the complete elliptic
C integral of the second kind (Hasting's approximation):
      m1=1.d0-k*k
      a1=0.44325141463d0
      a2=0.06260601220d0
      a3=0.04757383546d0
      a4=0.01736506451d0
      b1=0.24998368310d0
      b2=0.09200180037d0
      b3=0.04069697526d0
      b4=0.00526449639d0
      ee1=1.d0+m1*(a1+m1*(a2+m1*(a3+m1*a4)))
      ee2=m1*(b1+m1*(b2+m1*(b3+m1*b4)))*log(1.d0/m1)
      ellec=ee1+ee2
      return
      end

      function ellk(k)
      implicit none
      double precision a0,a1,a2,a3,a4,b0,b1,b2,b3,b4,ellk,
     &       ek1,ek2,k,m1
C Computes polynomial approximation for the complete elliptic
C integral of the first kind (Hasting's approximation):
      m1=1.d0-k*k
      a0=1.38629436112d0
      a1=0.09666344259d0
      a2=0.03590092383d0
      a3=0.03742563713d0
      a4=0.01451196212d0
      b0=0.5d0
      b1=0.12498593597d0
      b2=0.06880248576d0
      b3=0.03328355346d0
      b4=0.00441787012d0
      ek1=a0+m1*(a1+m1*(a2+m1*(a3+m1*a4)))
      ek2=(b0+m1*(b1+m1*(b2+m1*(b3+m1*b4))))*log(m1)
      ellk=ek1-ek2
      return
      end

      FUNCTION rf(x,y,z)
      REAL*8 rf,x,y,z,ERRTOL,TINY,BIG,THIRD,C1,C2,C3,C4
      PARAMETER (ERRTOL=.08d0,TINY=1.5d-38,BIG=3.d37,THIRD=1.d0/3.d0,
     *C1=1.d0/24.d0,C2=.1d0,C3=3.d0/44.d0,C4=1.d0/14.d0)
      REAL*8 alamb,ave,delx,dely,delz,e2,e3,sqrtx,sqrty,sqrtz,xt,yt,zt
      if(min(x,y,z).lt.0.d0.or.min(x+y,x+z,y+z).lt.TINY.or.max(x,y,
     *z).gt.BIG)pause 'invalid arguments in rf'
      xt=x
      yt=y
      zt=z
1     continue
        sqrtx=sqrt(xt)
        sqrty=sqrt(yt)
        sqrtz=sqrt(zt)
        alamb=sqrtx*(sqrty+sqrtz)+sqrty*sqrtz
        xt=.25d0*(xt+alamb)
        yt=.25d0*(yt+alamb)
        zt=.25d0*(zt+alamb)
        ave=THIRD*(xt+yt+zt)
        delx=(ave-xt)/ave
        dely=(ave-yt)/ave
        delz=(ave-zt)/ave
      if(max(abs(delx),abs(dely),abs(delz)).gt.ERRTOL)goto 1
      e2=delx*dely-delz**2
      e3=delx*dely*delz
      rf=(1.d0+(C1*e2-C2-C3*e3)*e2+C4*e3)/sqrt(ave)
      return
      END
C  (C) Copr. 1986-92 Numerical Recipes Software 0NL&WR2.
"""

import numpy as np
from numpy import sqrt, log
from PyAstronomy.pyaC import pyaErrors as PE

class OccultQuadPy:

    def occultquad(self,z0,u1,u2,p,nz=None):
        """
        Occultquad routine
        
        Python implementation of code provided by Mandel and Agol (2002).
        Please cite Mandel & Agol (2002) if you make use of this routine.
        The original authors are not responsible for the python
        re-implmentation.
        
        Parameters
        ----------
        z0 : array
            Impact parameter in units of stellar radius
        u1 : int
            Linear limb-darkening coefficient (gamma_1 in paper)
        u2 : int
            Quadratic limb-darkening coefficient (gamma_2 in paper)   
        p : int
            Occulting star (or planet) size in units of stellar radius
        nz : int, optional
            Number of values in z0. Determined here from z0, parameter
            introduced to maintain API.
        
        Returns
        -------
        muo1 : array
            Fraction of flux at each z0 (impact parameter) for a limb-darkened source
        mu0 : array
            Fraction of flux at each z0 (impact parameter) for a uniform source
        """
        # Determine number of impact parameter values (ignore keyword altogether)
        nz = len(z0)

        if abs(p-0.5) < 1e-3:
            p = 0.5

        omega = 1 - u1/3 - u2/6
        pi = np.pi
        lambdad = np.zeros(nz)
        etad = np.zeros(nz)
        lambdae = np.zeros(nz)
# C Loop over each impact parameter:
        for i in range(nz):
# C substitute z=z0(i) to shorten expressions
            z = z0[i]
            x1 = (p-z)**2
            x2 = (p+z)**2
            x3 = p**2-z**2
# C the source is unocculted:
# C Table 3, I.
            if z >= 1.0+p:
                lambdad[i] = 0
                etad[i] = 0
                lambdae[i] = 0
                continue
            elif (p >= 1.0) and (z <= p-1.0):
# C the  source is completely occulted:
# C Table 3, II.
                lambdad[i] = 1
                etad[i] = 1
                lambdae[i] = 1
                continue          
            elif (z >= abs(1-p)) and (z <= 1+p):
# C the source is partly occulted and the occulting object crosses the limb:
# C Equation (26):
                kap1 = np.arccos(min((1-p*p+z*z)/2/z,1))
                kap0 = np.arccos(min((p*p+z*z-1)/2/p/z,1))
                lambdae[i] = p*p*kap0+kap1
                lambdae[i] = (lambdae[i]-0.5*sqrt(max(4*z*z-(1+z*z-p*p)**2,0)))/pi

# C the occulting object transits the source star (but doesn't
# C completely cover it):
            if z <= 1-p:
                lambdae[i] = p*p
# C the edge of the occulting star lies at the origin- special 
# C expressions in this case:
            if abs(z-p) < 1e-4*(z+p):
# C Table 3, Case V.:
                if z >= 0.5:
                    lam = 0.5*pi
                    q = 0.5/p
                    Kk = self.ellk(q)
                    Ek = self.ellec(q)
# C Equation 34: lambda_3
                    lambdad[i] = 1/3+16*p/9/pi*(2*p*p-1)*Ek- \
                        (32*p**4-20*p*p+3)/9/pi/p*Kk
# C Equation 34: eta_1
                    etad[i] = 1/2/pi*(kap1+p*p*(p*p+2*z*z)*kap0- \
                                      (1+5*p*p+z*z)/4*sqrt((1-x1)*(x2-1)))
                    if p == 0.5:
# C Case VIII: p=1/2, z=1/2
                        lambdad[i] = 1/3-4/pi/9
                        etad[i] = 3/32
                    continue
                else:
# C Table 3, Case VI.:
                    lam = 0.5*pi
                    q = 2*p
                    Kk = self.ellk(q)
                    Ek = self.ellec(q)
# C Equation 34: lambda_4
                    lambdad[i] = 1/3+2/9/pi*(4*(2*p*p-1)*Ek + (1-4*p*p)*Kk)
# C Equation 34: eta_2
                    etad[i] = p*p/2*(p*p+2*z*z)
                    continue

# C the occulting star partly occults the source and crosses the limb:
# C Table 3, Case III:
            if ((z > 0.5+abs(p-0.5)) and (z < 1+p)) or ((p > 0.5) and (z > abs(1-p)*1.000001) and (z < p)):
                lam = 0.5*pi
                q = sqrt((1-(p-z)**2)/4/z/p)
                Kk = self.ellk(q)
                Ek = self.ellec(q)
                n = 1/x1-1
                Pk = Kk-n/3*self.rj(0,1-q*q,1,1+n)
# C Equation 34, lambda_1:
                lambdad[i] = 1/9/pi/sqrt(p*z)*(((1-x2)*(2*x2+ \
                                x1-3)-3*x3*(x2-2))*Kk+4*p*z*(z*z+ \
                                7*p*p-4)*Ek-3*x3/x1*Pk)
                if z < p:
                    lambdad[i] = lambdad[i]+2/3
# C Equation 34, eta_1:
                etad[i] = 1/2/pi*(kap1+p*p*(p*p+2*z*z)*kap0- \
                            (1+5*p*p+z*z)/4*sqrt((1-x1)*(x2-1)))
                continue
# C the occulting star transits the source:
# C Table 3, Case IV.:
            if (p <= 1) and (z <= (1-p)*1.000001):
                lam = 0.5*pi
                q = sqrt((x2-x1)/(1-x1))
                Kk = self.ellk(q)
                Ek = self.ellec(q)
                n = x2/x1-1
                Pk = Kk-n/3*self.rj(0,1-q*q,1,1+n)
# C Equation 34, lambda_2:
                lambdad[i] = 2/9/pi/sqrt(1-x1)*((1-5*z*z+p*p+ \
                                x3*x3)*Kk+(1-x1)*(z*z+7*p*p-4)*Ek-3*x3/x1*Pk)
                if z < p:
                    lambdad[i] = lambdad[i]+2/3
                if abs(p+z-1) <= 1e-4:
                    lambdad[i] = 2/3/pi*np.arccos(1-2*p)-4/9/pi*sqrt(p*(1-p))*(3+2*p-8*p*p)
# C Equation 34, eta_2:
                etad[i] = p*p/2*(p*p+2*z*z)
        
 # 10     continue
# C Now, using equation (33):
        
        muo1 = 1-((1-u1-2*u2)*lambdae+(u1+2*u2)*lambdad+u2*etad)/omega
# C Equation 25:
        mu0 = 1-lambdae
        return muo1, mu0


    def rf(self,x,y,z, ERRTOL=0.08, TINY=1.5e-38, BIG=3e37):
        if (min(x,y,z) < 0) or (min(x+y,x+z,y+z) < TINY) or (max(x,y,z) > BIG):
            raise(ValueError('invalid arguments in rf'))
        
        xt, yt, zt = x, y, z

        while True:
            sqrtx = sqrt(xt)
            sqrty = sqrt(yt)
            sqrtz = sqrt(zt)
            alamb = sqrtx*(sqrty+sqrtz) + sqrty*sqrtz
            xt = 0.25*(xt+alamb)
            yt = 0.25*(yt+alamb)
            zt = 0.25*(zt+alamb)
            ave = (xt+yt+zt)/3
            delx = (ave-xt)/ave
            dely = (ave-yt)/ave
            delz = (ave-zt)/ave
            if(max(abs(delx),abs(dely),abs(delz)) <= ERRTOL):
                break
     
        C1 = 1/24
        C2 = 0.1
        C3 = 3/44
        C4 = 1/14
     
        e2 = delx*dely-delz**2
        e3 = delx*dely*delz
        rf = (1.0+(C1*e2-C2-C3*e3)*e2+C4*e3)/sqrt(ave)
        return rf
      
    def rj(self,x,y,z,p,ERRTOL=0.05,TINY=2.5e-13,BIG=9.e11):

        if min(x,y,z) < 0:
            raise(PE.PyAValError(f"min(x,y,z) < 0 violated (x,y,z = {x}, {y}, {z})", \
                              where="rj in OccultQuadPy"))
        if min(x+y,x+z,y+z,abs(p)) < TINY:
            raise(PE.PyAValError(f"min(x+y,x+z,y+z,abs(p)) < TINY violated (x+y,x+z,y+z,abs(p) = {x+y}, {x+z}, {y+z}, {abs(p)})", \
                              where="rj in OccultQuadPy"))     
        if max(x,y,z,abs(p)) > BIG:
            raise(PE.PyAValError(f"max(x,y,z,abs(p) > BIG violated (x,y,z,abs(p) = {x}, {y}, {z}, {abs(p)})", \
                              where="rj in OccultQuadPy"))             
            
        
        summe = 0.0
        fac = 1.0
        if(p > 0.0):
            xt, yt, zt, pt = x, y, z, p
        else:
            xt = min(x,y,z)
            zt = max(x,y,z)
            yt = x+y+z-xt-zt
            a = 1.0/(yt-p)
            b = a*(zt-yt)*(yt-xt)
            pt = yt+b
            rho = xt*zt/yt
            tau = p*pt/yt
            rcx = self.rc(rho,tau)

        while True:
            sqrtx, sqrty, sqrtz = sqrt(xt), sqrt(yt), sqrt(zt)
            alamb = sqrtx*(sqrty+sqrtz)+sqrty*sqrtz
            alpha = (pt*(sqrtx+sqrty+sqrtz)+sqrtx*sqrty*sqrtz)**2
            beta = pt*(pt+alamb)**2
            summe += fac*self.rc(alpha,beta)
            fac = 0.25*fac
            xt = 0.25*(xt+alamb)
            yt = 0.25*(yt+alamb)
            zt = 0.25*(zt+alamb)
            pt = 0.25*(pt+alamb)
            ave = 0.2*(xt+yt+zt+pt+pt)
            delx=(ave-xt)/ave
            dely=(ave-yt)/ave
            delz=(ave-zt)/ave
            delp=(ave-pt)/ave
            if max(abs(delx),abs(dely),abs(delz),abs(delp)) <= ERRTOL:
                break
         
        ea = delx*(dely+delz)+dely*delz
        eb = delx*dely*delz 
        ec = delp**2
        ed = ea-3.0*ec
        ee = eb+2.0*delp*(ea-ec)
        
        C1=3/14
        C2=1/3
        C3=3/22
        C4=3/26
        C5=0.75*C3
        C6=1.5*C4
        C7=0.5*C2
        C8=C3+C3
        
        rj = 3.0*summe+fac*(1.0+ed*(-C1+C5*ed-C6*ee)+eb*(C7+delp*(-C8+delp*C4))+delp*ea*(C2-delp*C3)-C2*delp*ec)/(ave*sqrt(ave))
        if (p < 0):
            rj=a*(b*rj+3.0*(rcx-self.rf(xt,yt,zt)))
        return rj
      
    def rc(self,x,y,SQRTNY=1.3e-19,ERRTOL=0.04,TINY=1.69e-38,BIG=3e37):
        TNBG = TINY * BIG
        COMP1 = 2.236/SQRTNY
        COMP2 = TNBG*TNBG/25.0
        if (x < 0) or (y == 0) or (x+abs(y) < TINY) or (x+abs(y) > BIG) \
            or ( (y < -COMP1) and (x > 0) and (x < COMP2) ):
                raise(ValueError('invalid arguments in rc'))
        if y > 0:
            xt, yt, w = x, y, 1.
        else:
            xt = x-y
            yt = -y
            w = sqrt(x)/sqrt(xt)

        while True:
            alamb = 2.0*sqrt(xt)*sqrt(yt)+yt
            xt = 0.25*(xt+alamb)
            yt = 0.25*(yt+alamb)
            ave = (xt+yt+yt)/3
            s = (yt-ave)/ave
            if abs(s) <= ERRTOL:
                break
            
        C1 = 0.3
        C2 = 1.0/7.0
        C3 = 0.375
        C4 = 9.0/22
            
        rc = w*(1.0+s*s*(C1+s*(C2+s*(C3+s*C4))))/sqrt(ave)
        return rc

      
    def ellk(self, k):
        m1=1.0-k*k
        a0=1.38629436112
        a1=0.09666344259
        a2=0.03590092383
        a3=0.03742563713
        a4=0.01451196212
        b0=0.5
        b1=0.12498593597
        b2=0.06880248576
        b3=0.03328355346
        b4=0.00441787012
        ek1 = a0+m1*(a1+m1*(a2+m1*(a3+m1*a4)))
        ek2 = (b0+m1*(b1+m1*(b2+m1*(b3+m1*b4))))*log(m1)
        ellk = ek1-ek2
        return ellk
    
    def ellec(self, k):
        m1=1.0-k*k
        a1=0.44325141463
        a2=0.06260601220
        a3=0.04757383546
        a4=0.01736506451
        b1=0.24998368310
        b2=0.09200180037
        b3=0.04069697526
        b4=0.00526449639
        ee1 = 1.0+m1*(a1+m1*(a2+m1*(a3+m1*a4)))
        ee2 = m1*(b1+m1*(b2+m1*(b3+m1*b4)))*log(1.0/m1)
        ellec = ee1+ee2
        return ellec


      


# -*- coding: utf-8 -*-
from __future__ import print_function, division
_mplImport = True
try:
  import matplotlib.pylab as mpl
except ImportError:
  _mplImport = False
import numpy
from PyAstronomy.pyaC import pyaErrors as PE

class PeriodBase:
  """
    Base class for all periodograms.
  
    This class provides the framework
    for all periodograms within *pyPeriod* package.

    *PeriodBase* has a *plot* method, which can be used to provide
    a quick-look of the result. The significance of a
    feature with a power *Pn* can be assessed using the *prob* and
    *FAP* methods.

    Attributes
    ----------
    power : array
        The periodogram power.
    freq : array
        The frequencies at which the power are evaluated.
  """

  def __calcPeriodogram(self):
    pass

  def prob(self,Pn):
    raise(PE.PyANotImplemented("`prob` is not yet implemented."))

  def plot(self, *args,**kwargs):
    """
      Creates a matplotlib figure and axes class instance to visualize the result.
      
      Parameters:
        - `FAPlevels` - optional, List of false-alarm probability (FAP) levels
        - `*args` - optional, Arguments passed to plot method of axes class.
        - `**kwargs` - optional, Keyword arguments passed plot method to axes class.
      
      This method provides a quick and simple way to visualize the results of the a
      periodogram calculation.
      
      Returns:
        The created *Figure* and *Axes* class instances.
    """
    if not _mplImport:
      raise(PE.PyARequiredImport("To use this function matplotlib must be installed.", \
                                 where="PeriodBase::plot", solution=["Install matplotlib.", "Avoid call and use custom plotting interface."]))
    fig = mpl.figure()
    ax = fig.add_subplot(1,1,1)

    if "FAPlevels" in kwargs:
      fapLvs = numpy.array(kwargs["FAPlevels"])
      del kwargs["FAPlevels"]
      powLvs = self.powerLevel(fapLvs)
      for i,powLv in enumerate(powLvs):
        ax.plot([self.freq.min(),self.freq.max()],[powLv, powLv],'k--')
        ax.annotate("FAP "+("%2.3g" % fapLvs[i]), xy=(self.freq.max(),powLv), xytext=None, xycoords='data',
           textcoords='data', arrowprops=None, horizontalalignment="right")
      
    if self.label is None:
      ax.set_title("Periodogram")
      ax.set_xlabel("Frequency")
      ax.set_ylabel("Power")
    else:
      ax.set_title(self.label["title"])
      ax.set_xlabel(self.label["xlabel"])
      ax.set_ylabel(self.label["ylabel"])
    ax.plot(self.freq, self.power, *args, **kwargs)

    return fig, ax

  def FAP(self, Pn):
    """
      Obtain the false-alarm probability (FAP).

      The FAP denotes the probability that at least one out of M
      independent power values in a prescribed search band of a
      power spectrum computed from a white-noise time series is
      as large as or larger than the threshold, `Pn`.
      It is assessed through
      
      .. math:: FAP(Pn) = 1 - (1-Prob(P>Pn))^M \\; ,
      
      where "Prob(P>Pn)" depends on the type of periodogram
      and normalization and is
      calculated by using the *prob* method;
      *M* is the number of independent power
      values and is computed internally.

      Parameters
      ----------
      Pn : float
          Power threshold.
      
      Returns
      -------
      FAP : float
          False alarm probability.
    """
    prob = self.M * self.prob(Pn)
    if prob > 0.01:  return 1.-(1.-self.prob(Pn))**self.M
    return prob
  
  def probInv(self, Prob):
    raise(PE.PyANotImplemented("`probInv` is not yet implemented."))
  
  def powerLevel(self, FAPlevel):
    """
      Power threshold for FAP level.
    
      Parameters
      ----------
      FAPlevel : float or array
            "False Alarm Probability" threshold
    
      Returns
      -------
      Threshold : float or array
          The power threshold pertaining to a specified
          false-alarm probability (FAP). Powers exceeding this
          threshold have FAPs smaller than FAPlevel.
    """
    Prob = 1.-(1.-FAPlevel)**(1./self.M)
    return self.probInv(Prob)
    
  
  def stats(self, Pn):
    """
      Obtain basic statistics for power threshold.
      
      Parameters
      ----------
      Pn : float
          Power threshold.
    
      Returns
      -------
      Statistics : dictionary
          A dictionary containing {'Pn': *Pn*, 'FAP': *FAP(Pn)* ,
          'Prob': *Prob(Pn)*} for the specified power threshold, *Pn*.
    """
    return {'Pn': Pn, 'Prob': self.prob(Pn), 'FAP': self.FAP(Pn)}




class TimeSeries:
  """
    A container class that holds the observed light curve.
    
    Parameters
    ----------
    time : array
        The time array.
    flux : array
        The observed flux/data.
    error : array, optional
        The error of the data values.
  """

  def __init__(self, time, flux, error=None):
    self.time = numpy.array(time)
    self.flux = numpy.array(flux)
    self.error = error

    if len(self.time) != len(self.flux):
      raise(PE.PyAValError("Incompatible dimensions of input data arrays (time and flux).", \
                           where="TimeSeries::__init__", solution="Check the input arrays!"))

  def returnNyquist(self):
    """
      Calculate the average Nyquist frequency.
      
      Returns
      -------
      Nyquist frequency : float
          Half the sampling frequency of the time series.
    """
    return 0.5*1./numpy.mean(self.time[1::]-self.time[0:-1])